import polars as pl
from abc import abstractmethod
from typing import Any, Self

from ziplime.data.domain.data_bundle import DataBundle


class BundleStorage:

    def __init__(self):
        pass

    @abstractmethod
    async def store_bundle(self, data_bundle: DataBundle): ...

    @abstractmethod
    async def load_data_bundle(self, data_bundle: DataBundle) -> pl.DataFrame: ...

    @classmethod
    @abstractmethod
    async def from_json(cls, data: dict[str, Any]) -> Self: ...

    @abstractmethod
    async def to_json(self, data_bundle: DataBundle) -> dict[str, Any]: ...



import datetime
import logging
import multiprocessing

import limexhub
import numpy
import pandas as pd
from joblib import Parallel, delayed

from ziplime.constants.fundamental_data import FundamentalData
from ziplime.data.abstract_fundamendal_data_provider import AbstractFundamentalDataProvider


def _get_fundamental_data_task(
        symbol: str,
        date_from: datetime.datetime,
        date_to: datetime.datetime,
        frequency: datetime.timedelta,
        fundamental_data_list: set[str],
        limex_api_key: str
):
    # print("FUNDAMENTAL DATA")
    limex_client = limexhub.RestAPI(token=limex_api_key)

    dr = pd.date_range(date_from, date_to, freq='D')

    if fundamental_data_list:
        fundamental = limex_client.fundamental(
            symbol=symbol,
            from_date=date_from,
            to_date=date_to,
            fields=None
        )
        if fundamental.empty:
            logging.warning(f"No fundamental data for {symbol} for period {date_from}-{date_to}.")
    else:
        fundamental = pd.DataFrame(columns=["date", "symbol"])
        fundamental.set_index("date", inplace=True, drop=True)
        fundamental = fundamental.reindex(dr, fill_value=None)
        fundamental["symbol"] = symbol
        fundamental.reset_index(inplace=True)
        fundamental.rename(columns={"index": "date"}, inplace=True)
        fundamental.reset_index(inplace=True)
        return fundamental
    if not fundamental.empty:
        fundamental["date"] = pd.to_datetime(fundamental.date, utc=True)
    fundamental_new = pd.DataFrame(columns=["date", "symbol"])
    fundamental_new.set_index(keys=["date", "symbol"], inplace=True, drop=True)
    use_datetime = True
    for fund_col in FundamentalData:

        # print(fund_col.value)

        col_name = fund_col.value
        ttm_col = f"{col_name}_ttm"
        value_col = f"{col_name}_value"
        add_value_col = f"{col_name}_add_value"
        columns = ["date", "symbol"]
        if col_name in fundamental_data_list:
            columns.append(ttm_col)
            columns.append(value_col)
            columns.append(add_value_col)
        else:
            if ttm_col in fundamental_data_list:
                columns.append(ttm_col)
            if value_col in fundamental_data_list:
                columns.append(value_col)
            if add_value_col in fundamental_data_list:
                columns.append(add_value_col)

        res_df = pd.DataFrame(columns=columns)
        if len(columns) <= 2:
            continue
        if "field" in fundamental.columns:
            values_for_col = fundamental[fundamental.field == col_name]
            for index, row in values_for_col.iterrows():
                if use_datetime:
                    dt = datetime.datetime.combine(row.date, time=datetime.time(), tzinfo=datetime.timezone.utc)
                else:
                    dt = row.date

                vals = [dt, symbol]
                if ttm_col in columns:
                    vals.append(row.ttm)
                if value_col in columns:
                    vals.append(row.value)
                if add_value_col in columns:
                    vals.append(row.add_value)
                vals_df = pd.DataFrame([
                    vals,
                ], columns=res_df.columns)
                res_df = (
                    res_df.copy()
                    if vals_df.empty
                    else vals_df
                    if res_df.empty
                    else pd.concat([res_df, vals_df], ignore_index=True)
                )
        else:
            vals = [date_from, symbol]
            if ttm_col in columns:
                vals.append(numpy.nan)
            if value_col in columns:
                vals.append(numpy.nan)
            if add_value_col in columns:
                vals.append(numpy.nan)
            vals_df = pd.DataFrame([
                vals,
            ], columns=res_df.columns)

            res_df = (
                res_df.copy()
                if vals_df.empty
                else vals_df
                if res_df.empty
                else pd.concat([vals_df, res_df], ignore_index=True)
            )
        res_df = res_df.drop_duplicates(subset=['date', 'symbol'], keep='last')
        res_df.set_index("date", inplace=True, drop=True)
        res_df = res_df.reindex(dr, fill_value=None)
        res_df["symbol"] = symbol
        res_df.reset_index(inplace=True)
        res_df.rename(columns={"index": "date"}, inplace=True)
        res_df.set_index(["date", "symbol"], inplace=True)
        fundamental_new = pd.concat([res_df, fundamental_new], ignore_index=False, axis=1)

    fundamental_new.reset_index(inplace=True)
    filtered = fundamental_new[fundamental_new.symbol.notnull()]
    # filtered.set_index(["date", "symbol"], inplace=True)

    return filtered


class LimexHubFundamentalDataProvider(AbstractFundamentalDataProvider):
    def __init__(self, limex_api_key: str, maximum_threads: int | None = None):
        self._limex_api_key = limex_api_key
        self._logger = logging.getLogger(__name__)
        self._limex_client = limexhub.RestAPI(token=limex_api_key)
        if maximum_threads is not None:
            self._maximum_threads = min(multiprocessing.cpu_count() * 2, maximum_threads)
        else:
            self._maximum_threads = multiprocessing.cpu_count() * 2

    def get_fundamental_data(self,
                             symbols: list[str],
                             date_from: datetime.datetime,
                             date_to: datetime.datetime,
                             frequency: datetime.timedelta,
                             fundamental_data_list: set[str]
                             ):

        def fetch_fundamental(limex_api_key: str, symbol: str):
            try:
                result = _get_fundamental_data_task(date_from=date_from, date_to=date_to,
                                                    limex_api_key=limex_api_key, symbol=symbol, frequency=frequency,
                                                    fundamental_data_list=fundamental_data_list)
                return result
            except Exception as e:
                logging.exception(
                    f"Exception fetching fundamental data for symbol {symbol}, date_from={date_from}, date_to={date_to}. Skipping."
                )
                return None

        data = pd.DataFrame()

        res = Parallel(n_jobs=self._maximum_threads, prefer="threads",
                       return_as="generator_unordered")(
            delayed(fetch_fundamental)(self._limex_api_key, symbol) for symbol in symbols)
        for item in res:
            if item is None:
                continue
            data = pd.concat([data, item], axis=0, ignore_index=False)
        data.set_index(["date", "symbol"], inplace=True)

        return data

    def get_fundamental_data_column_names(self, fundamental_data_fields: frozenset[str]) -> set[str]:
        columns = set()
        for fund_col in FundamentalData:
            col_name = fund_col.value
            ttm_col = f"{col_name}_ttm"
            value_col = f"{col_name}_value"
            add_value_col = f"{col_name}_add_value"
            if col_name in fundamental_data_fields:
                columns.add(ttm_col)
                columns.add(value_col)
                columns.add(add_value_col)
            else:
                if ttm_col in fundamental_data_fields:
                    columns.add(ttm_col)
                if value_col in fundamental_data_fields:
                    columns.add(value_col)
                if add_value_col in fundamental_data_fields:
                    columns.add(add_value_col)

        return columns



import datetime
from dataclasses import dataclass, field
from decimal import Decimal

from ziplime.assets.entities.asset import Asset
from ziplime.domain.position import Position


@dataclass
class Portfolio:
    # capital_used: float
    cash_flow: Decimal
    starting_cash: Decimal
    portfolio_value: Decimal
    pnl: Decimal
    returns: Decimal
    cash: Decimal
    positions_value: Decimal
    positions_exposure: Decimal
    positions: dict[Asset, Position] = field(default_factory=dict)
    start_date: datetime.datetime | None = None





from decimal import Decimal

from ziplime.trading.entities.orders.order_request import OrderRequest
from ziplime.trading.enums.order_side import OrderSide
from ziplime.trading.enums.order_type import OrderType


class StopLimitOrder(OrderRequest):

    def __init__(self, order_id: str, order_side: OrderSide, quantity: Decimal, limit_price: Decimal,
               stop_price: Decimal):
        super().__init__(order_id=order_id, order_type=OrderType.LIMIT)
        self.order_side = order_side
        self.quantity = quantity
        self.limit_price = limit_price
        self.stop_price = stop_price



import aiofiles.os
from pathlib import Path
from typing import Any, Sequence, Literal, Self
from polars import CredentialProviderFunction
from polars._typing import ParquetCompression
import polars as pl
from ziplime.data.domain.data_bundle import DataBundle
from ziplime.data.services.bundle_storage import BundleStorage


class FileSystemParquetBundleStorage(BundleStorage):

    def __init__(self,
                 base_data_path: str,
                 compression: ParquetCompression = "brotli",
                 compression_level: int | None = None,
                 statistics: bool | str | dict[str, bool] = True,
                 row_group_size: int | None = None,
                 data_page_size: int | None = None,
                 use_pyarrow: bool = False,
                 pyarrow_options: dict[str, Any] | None = None,
                 partition_by: str | Sequence[str] | None = None,
                 partition_chunk_size_bytes: int = 4_294_967_296,
                 storage_options: dict[str, Any] | None = None,
                 credential_provider: (
                         CredentialProviderFunction | Literal["auto"] | None
                 ) = "auto", ):
        super().__init__()
        self.base_data_path = base_data_path
        self.compression = compression
        self.compression_level = compression_level
        self.statistics = statistics
        self.row_group_size = row_group_size
        self.data_page_size = data_page_size
        self.use_pyarrow = use_pyarrow
        self.pyarrow_options = pyarrow_options
        self.partition_by = partition_by
        self.partition_chunk_size_bytes = partition_chunk_size_bytes
        self.storage_options = storage_options

    async def store_bundle(self, data_bundle: DataBundle):
        # we need here to know ehere to store bundle, and info is in bundle metadata
        bundle_path = self.get_data_bundle_path(data_bundle=data_bundle)
        await aiofiles.os.makedirs(bundle_path.parent, exist_ok=True)
        data_bundle.data.write_parquet(bundle_path, compression=self.compression,
                                       compression_level=10,
                                       statistics=self.statistics, row_group_size=self.row_group_size,
                                       data_page_size=self.data_page_size, use_pyarrow=self.use_pyarrow,
                                       pyarrow_options=self.pyarrow_options, partition_by=self.partition_by,
                                       partition_chunk_size_bytes=self.partition_chunk_size_bytes,
                                       storage_options=self.storage_options)

    async def load_data_bundle(self, data_bundle: DataBundle) -> pl.DataFrame:
        bundle_path = self.get_data_bundle_path(data_bundle=data_bundle)
        data = pl.read_parquet(source=bundle_path)
        return data


    @classmethod
    async def from_json(cls, data: dict[str, Any]) -> Self:
        return cls(base_data_path=data["base_data_path"])

    def get_data_bundle_path(self, data_bundle: DataBundle) -> Path:
        return Path(self.base_data_path, "data_bundle", data_bundle.name, data_bundle.version, f"data.parquet")

    async def to_json(self, data_bundle: DataBundle) -> dict[str, Any]:
        return {
            "base_data_path": self.base_data_path,
        }



ALL_US_STOCK_SYMBOLS = ["A", "AA", "AACB", "AACBR", "AACBU", "AACG", "AACT", "AAL", "AAM", "AAME", "AAMI", "AAOI", "AAON",
                     "AAP", "AAPG", "AAPL", "AARD", "AAT", "AB", "ABAT", "ABBV", "ABCB", "ABCL", "ABEO", "ABEV", "ABG",
                     "ABL", "ABLLL", "ABLLW", "ABLV", "ABM", "ABNB", "ABOS", "ABP", "ABPWW", "ABR", "ABSI", "ABT",
                     "ABTS", "ABUS", "ABVC", "ABVE", "ABVEW", "ABVX", "AC", "ACA", "ACAD", "ACB", "ACCO", "ACCS",
                     "ACDC", "ACEL", "ACET", "ACGL", "ACGLN", "ACGLO", "ACHC", "ACHR", "ACHV", "ACI", "ACIC", "ACIU",
                     "ACIW", "ACLS", "ACLX", "ACM", "ACMR", "ACN", "ACNB", "ACNT", "ACOG", "ACON", "ACONW", "ACP",
                     "ACR", "ACRE", "ACRS", "ACRV", "ACT", "ACTG", "ACTU", "ACU", "ACV", "ACVA", "ACXP", "ADAG", "ADAP",
                     "ADBE", "ADC", "ADCT", "ADD", "ADEA", "ADGM", "ADI", "ADIL", "ADM", "ADMA", "ADN", "ADNT", "ADNWW",
                     "ADP", "ADPT", "ADSE", "ADSEW", "ADSK", "ADT", "ADTN", "ADTX", "ADUR", "ADUS", "ADV", "ADVB",
                     "ADVM", "ADVWW", "ADX", "ADXN", "AEE", "AEF", "AEFC", "AEG", "AEHL", "AEHR", "AEI", "AEIS", "AEM",
                     "AEMD", "AENT", "AENTW", "AEO", "AEON", "AEP", "AER", "AERT", "AERTW", "AES", "AESI", "AEVA",
                     "AEVAW", "AEYE", "AFB", "AFBI", "AFCG", "AFG", "AFGB", "AFGC", "AFGD", "AFGE", "AFJKR", "AFL",
                     "AFMD", "AFRI", "AFRIW", "AFRM", "AFYA", "AG", "AGAE", "AGCO", "AGD", "AGEN", "AGFY", "AGH", "AGI",
                     "AGIO", "AGL", "AGM", "AGMH", "AGNC", "AGNCL", "AGNCM", "AGNCN", "AGNCO", "AGNCP", "AGO", "AGRI",
                     "AGRO", "AGS", "AGX", "AGYS", "AHCO", "AHG", "AHH", "AHL", "AHR", "AHT", "AI", "AIFE", "AIFF",
                     "AIFU", "AIG", "AIHS", "AIMAU", "AIMAW", "AIMBU", "AIMD", "AIMDW", "AIN", "AIO", "AIOT", "AIP",
                     "AIR", "AIRE", "AIRG", "AIRI", "AIRJ", "AIRJW", "AIRS", "AIRT", "AIRTP", "AISP", "AISPW", "AIT",
                     "AIV", "AIXI", "AIZ", "AIZN", "AJG", "AKA", "AKAM", "AKAN", "AKBA", "AKO", "AKR", "AKRO", "AKTX",
                     "AKYA", "AL", "ALAB", "ALAR", "ALB", "ALBT", "ALC", "ALCO", "ALCY", "ALCYW", "ALDF", "ALDFU",
                     "ALDFW", "ALDX", "ALE", "ALEC", "ALEX", "ALF", "ALFUU", "ALFUW", "ALG", "ALGM", "ALGN", "ALGS",
                     "ALGT", "ALHC", "ALIT", "ALK", "ALKS", "ALKT", "ALL", "ALLE", "ALLK", "ALLO", "ALLR", "ALLT",
                     "ALLY", "ALMS", "ALMU", "ALNT", "ALNY", "ALOT", "ALRM", "ALRS", "ALSN", "ALT", "ALTG", "ALTI",
                     "ALTO", "ALTS", "ALUR", "ALV", "ALVO", "ALVOW", "ALX", "ALXO", "ALZN", "AM", "AMAL", "AMAT",
                     "AMBA", "AMBC", "AMBI", "AMBO", "AMBP", "AMBR", "AMC", "AMCR", "AMCX", "AMD", "AME", "AMED", "AMG",
                     "AMGN", "AMH", "AMIX", "AMKR", "AMLX", "AMN", "AMOD", "AMODW", "AMP", "AMPG", "AMPGW", "AMPH",
                     "AMPL", "AMPX", "AMPY", "AMR", "AMRC", "AMRK", "AMRN", "AMRX", "AMS", "AMSC", "AMSF", "AMST",
                     "AMT", "AMTB", "AMTD", "AMTM", "AMTX", "AMWD", "AMWL", "AMX", "AMZE", "AMZN", "AN", "ANAB", "ANDE",
                     "ANEB", "ANET", "ANF", "ANG", "ANGH", "ANGHW", "ANGI", "ANGO", "ANIK", "ANIP", "ANIX", "ANL",
                     "ANNA", "ANNAW", "ANNX", "ANRO", "ANSC", "ANSCW", "ANSS", "ANTE", "ANTX", "ANVS", "ANY", "AOD",
                     "AOMN", "AOMR", "AON", "AORT", "AOS", "AOSL", "AOUT", "AP", "APA", "APAM", "APCX", "APD", "APDN",
                     "APEI", "APG", "APGE", "APH", "API", "APLD", "APLE", "APLM", "APLS", "APLT", "APM", "APO", "APOG",
                     "APOS", "APP", "APPF", "APPN", "APPS", "APRE", "APT", "APTV", "APVO", "APWC", "APYX", "AQB",
                     "AQMS", "AQN", "AQNB", "AQST", "AR", "ARAY", "ARBB", "ARBE", "ARBEW", "ARBK", "ARBKL", "ARCB",
                     "ARCC", "ARCO", "ARCT", "ARDC", "ARDT", "ARDX", "ARE", "AREB", "AREBW", "AREC", "AREN", "ARES",
                     "ARGD", "ARGO", "ARGX", "ARHS", "ARI", "ARIS", "ARKO", "ARKOW", "ARKR", "ARL", "ARLO", "ARLP",
                     "ARM", "ARMK", "ARMN", "ARMP", "AROC", "AROW", "ARQ", "ARQQ", "ARQQW", "ARQT", "ARR", "ARRY",
                     "ARTL", "ARTNA", "ARTV", "ARTW", "ARVN", "ARW", "ARWR", "AS", "ASA", "ASAN", "ASB", "ASBA", "ASBP",
                     "ASBPW", "ASC", "ASG", "ASGI", "ASGN", "ASH", "ASIX", "ASLE", "ASM", "ASMB", "ASML", "ASND",
                     "ASNS", "ASO", "ASPC", "ASPI", "ASPN", "ASPS", "ASR", "ASRT", "ASRV", "ASST", "ASTC", "ASTE",
                     "ASTH", "ASTI", "ASTL", "ASTLW", "ASTS", "ASUR", "ASX", "ASYS", "ATAI", "ATAT", "ATCH", "ATCO",
                     "ATCOL", "ATEC", "ATEN", "ATER", "ATEX", "ATGE", "ATGL", "ATH", "ATHA", "ATHE", "ATHM", "ATHR",
                     "ATHS", "ATI", "ATII", "ATIIU", "ATIIW", "ATKR", "ATLC", "ATLCL", "ATLCP", "ATLCZ", "ATLN", "ATLO",
                     "ATLX", "ATMC", "ATMU", "ATMVR", "ATNF", "ATNFW", "ATNI", "ATNM", "ATO", "ATOM", "ATOS", "ATPC",
                     "ATR", "ATRA", "ATRC", "ATRO", "ATS", "ATUS", "ATXG", "ATXS", "ATYR", "AU", "AUB", "AUBN", "AUDC",
                     "AUID", "AUNA", "AUPH", "AUR", "AURA", "AUROW", "AUST", "AUTL", "AUUD", "AUUDW", "AVA", "AVAH",
                     "AVAL", "AVAV", "AVB", "AVBP", "AVD", "AVDL", "AVDX", "AVGO", "AVIR", "AVK", "AVNS", "AVNT",
                     "AVNW", "AVO", "AVPT", "AVPTW", "AVR", "AVT", "AVTE", "AVTR", "AVTX", "AVXL", "AVY", "AWF", "AWI",
                     "AWK", "AWP", "AWR", "AWRE", "AWX", "AX", "AXDX", "AXGN", "AXIL", "AXL", "AXON", "AXP", "AXR",
                     "AXS", "AXSM", "AXTA", "AXTI", "AYI", "AYRO", "AYTU", "AZ", "AZEK", "AZI", "AZN", "AZO", "AZTA",
                     "AZTR", "AZUL", "AZZ", "B", "BA", "BABA", "BAC", "BACQ", "BACQU", "BAER", "BAERW", "BAFN", "BAH",
                     "BAK", "BALL", "BALY", "BAM", "BANC", "BAND", "BANF", "BANFP", "BANL", "BANR", "BANX", "BAOS",
                     "BAP", "BARK", "BASE", "BATL", "BATRA", "BATRK", "BAX", "BAYA", "BAYAR", "BB", "BBAI", "BBAR",
                     "BBCP", "BBD", "BBDC", "BBDO", "BBGI", "BBIO", "BBLG", "BBLGW", "BBN", "BBNX", "BBSI", "BBU",
                     "BBUC", "BBVA", "BBW", "BBWI", "BBY", "BC", "BCAB", "BCAL", "BCAT", "BCAX", "BCBP", "BCC", "BCDA",
                     "BCE", "BCG", "BCH", "BCLI", "BCML", "BCO", "BCPC", "BCRX", "BCS", "BCSF", "BCTX", "BCTXW", "BCV",
                     "BCX", "BCYC", "BDC", "BDJ", "BDL", "BDMD", "BDN", "BDRX", "BDSX", "BDTX", "BDX", "BE", "BEAG",
                     "BEAGR", "BEAGU", "BEAM", "BEAT", "BEATW", "BECN", "BEDU", "BEEM", "BEEP", "BEKE", "BELFA",
                     "BELFB", "BEN", "BENF", "BENFW", "BEP", "BEPC", "BEPH", "BEPI", "BEPJ", "BERY", "BETR", "BF",
                     "BFAM", "BFC", "BFH", "BFIN", "BFK", "BFLY", "BFRG", "BFRGW", "BFRI", "BFRIW", "BFS", "BFST",
                     "BFZ", "BG", "BGB", "BGC", "BGFV", "BGH", "BGI", "BGLC", "BGM", "BGR", "BGS", "BGSF", "BGT", "BGX",
                     "BGY", "BH", "BHAT", "BHB", "BHC", "BHE", "BHF", "BHFAL", "BHFAM", "BHFAN", "BHFAO", "BHFAP",
                     "BHK", "BHLB", "BHM", "BHP", "BHR", "BHRB", "BHST", "BHV", "BHVN", "BIAF", "BIDU", "BIGC", "BIIB",
                     "BILI", "BILL", "BIO", "BIOA", "BIOX", "BIP", "BIPC", "BIPH", "BIPI", "BIPJ", "BIRD", "BIRK",
                     "BIT", "BITF", "BIVI", "BIYA", "BJ", "BJDX", "BJRI", "BK", "BKD", "BKE", "BKH", "BKHA", "BKHAR",
                     "BKKT", "BKN", "BKNG", "BKR", "BKSY", "BKT", "BKTI", "BKU", "BKV", "BKYI", "BL", "BLBD", "BLBX",
                     "BLCO", "BLD", "BLDE", "BLDEW", "BLDP", "BLDR", "BLE", "BLFS", "BLFY", "BLIN", "BLIV", "BLK",
                     "BLKB", "BLMN", "BLMZ", "BLND", "BLNE", "BLNK", "BLRX", "BLTE", "BLUE", "BLW", "BLX", "BLZE",
                     "BMA", "BMBL", "BME", "BMEA", "BMEZ", "BMGL", "BMI", "BML", "BMN", "BMO", "BMR", "BMRA", "BMRC",
                     "BMRN", "BMY", "BN", "BNAI", "BNAIW", "BNED", "BNGO", "BNH", "BNJ", "BNL", "BNR", "BNRG", "BNS",
                     "BNT", "BNTC", "BNTX", "BNY", "BNZI", "BNZIW", "BOC", "BODI", "BOE", "BOF", "BOH", "BOKF", "BOLD",
                     "BOLT", "BON", "BOOM", "BOOT", "BORR", "BOSC", "BOTJ", "BOW", "BOWN", "BOWNR", "BOWNU", "BOX",
                     "BOXL", "BP", "BPMC", "BPOP", "BPOPM", "BPRN", "BPT", "BPYPM", "BPYPN", "BPYPO", "BPYPP", "BQ",
                     "BR", "BRAG", "BRBR", "BRBS", "BRC", "BRCC", "BRDG", "BREA", "BRFH", "BRFS", "BRIA", "BRID", "BRK",
                     "BRKL", "BRKR", "BRLS", "BRLSW", "BRLT", "BRN", "BRNS", "BRO", "BROG", "BROS", "BRSP", "BRT",
                     "BRTX", "BRW", "BRX", "BRY", "BRZE", "BSAC", "BSBK", "BSBR", "BSET", "BSGM", "BSII", "BSIIW",
                     "BSL", "BSLK", "BSM", "BSRR", "BST", "BSTZ", "BSVN", "BSX", "BSY", "BTA", "BTAI", "BTBD", "BTBT",
                     "BTCM", "BTCS", "BTCT", "BTDR", "BTE", "BTG", "BTI", "BTM", "BTMD", "BTMWW", "BTO", "BTOC", "BTOG",
                     "BTSG", "BTSGU", "BTT", "BTU", "BTX", "BTZ", "BUD", "BUI", "BUJA", "BUJAR", "BUJAU", "BUJAW",
                     "BULL", "BULLW", "BULLZ", "BUR", "BURL", "BURU", "BUSE", "BV", "BVFL", "BVN", "BVS", "BW", "BWA",
                     "BWAY", "BWB", "BWBBP", "BWEN", "BWFG", "BWG", "BWIN", "BWLP", "BWMN", "BWMX", "BWNB", "BWSN",
                     "BWXT", "BX", "BXC", "BXMT", "BXMX", "BXP", "BXSL", "BY", "BYD", "BYFC", "BYM", "BYND", "BYON",
                     "BYRN", "BYSI", "BZ", "BZAI", "BZAIW", "BZFD", "BZFDW", "BZH", "BZUN", "C", "CAAP", "CAAS", "CABA",
                     "CABO", "CAC", "CACC", "CACI", "CADE", "CADL", "CAE", "CAF", "CAG", "CAH", "CAKE", "CAL", "CALC",
                     "CALM", "CALX", "CAMP", "CAMT", "CAN", "CANF", "CANG", "CAPL", "CAPN", "CAPR", "CAPS", "CAPT",
                     "CAR", "CARE", "CARG", "CARM", "CARR", "CARS", "CART", "CARV", "CASH", "CASI", "CASK", "CASS",
                     "CASY", "CAT", "CATO", "CATX", "CATY", "CAVA", "CB", "CBAN", "CBAT", "CBFV", "CBL", "CBLL", "CBNA",
                     "CBNK", "CBRE", "CBRL", "CBSH", "CBT", "CBU", "CBUS", "CBZ", "CC", "CCAP", "CCB", "CCBG", "CCCC",
                     "CCCS", "CCD", "CCEC", "CCEL", "CCEP", "CCG", "CCGWW", "CCI", "CCIA", "CCIF", "CCIR", "CCIRU",
                     "CCIRW", "CCIX", "CCIXU", "CCIXW", "CCJ", "CCK", "CCL", "CCLD", "CCLDO", "CCM", "CCNE", "CCNEP",
                     "CCO", "CCOI", "CCRD", "CCRN", "CCS", "CCSI", "CCTG", "CCU", "CCZ", "CDE", "CDIO", "CDIOW", "CDLR",
                     "CDLX", "CDNA", "CDNS", "CDP", "CDR", "CDRE", "CDRO", "CDROW", "CDT", "CDTG", "CDTTW", "CDTX",
                     "CDW", "CDXS", "CDZI", "CDZIP", "CE", "CEAD", "CEADW", "CECO", "CEE", "CEG", "CELC", "CELH",
                     "CELU", "CELUW", "CELZ", "CENN", "CENT", "CENTA", "CENX", "CEP", "CEPO", "CEPU", "CERO", "CEROW",
                     "CERS", "CERT", "CET", "CETX", "CETY", "CEV", "CEVA", "CF", "CFBK", "CFFI", "CFFN", "CFG", "CFLT",
                     "CFR", "CFSB", "CG", "CGABL", "CGAU", "CGBD", "CGBDL", "CGBS", "CGBSW", "CGC", "CGEM", "CGEN",
                     "CGNT", "CGNX", "CGO", "CGON", "CGTL", "CGTX", "CHA", "CHAR", "CHARR", "CHCI", "CHCO", "CHCT",
                     "CHD", "CHDN", "CHE", "CHEB", "CHEF", "CHEK", "CHGG", "CHH", "CHI", "CHKP", "CHMG", "CHMI", "CHN",
                     "CHNR", "CHPT", "CHR", "CHRD", "CHRO", "CHRS", "CHRW", "CHSCL", "CHSCM", "CHSCN", "CHSCO", "CHSCP",
                     "CHSN", "CHT", "CHTR", "CHW", "CHWY", "CHX", "CHY", "CI", "CIA", "CIB", "CICB", "CIEN", "CIF",
                     "CIFR", "CIFRW", "CIG", "CIGI", "CIGL", "CII", "CIIT", "CIK", "CIM", "CIMN", "CIMO", "CINF",
                     "CING", "CINGW", "CINT", "CIO", "CION", "CISO", "CISS", "CIVB", "CIVI", "CIX", "CJET", "CJMB",
                     "CKPT", "CKX", "CL", "CLAR", "CLB", "CLBK", "CLBR", "CLBT", "CLCO", "CLDI", "CLDT", "CLDX", "CLEU",
                     "CLF", "CLFD", "CLGN", "CLH", "CLIK", "CLIR", "CLLS", "CLM", "CLMB", "CLMT", "CLNE", "CLNN",
                     "CLNNW", "CLOV", "CLPR", "CLPS", "CLPT", "CLRB", "CLRO", "CLS", "CLSD", "CLSK", "CLSKW", "CLST",
                     "CLVT", "CLW", "CLWT", "CLX", "CLYM", "CM", "CMA", "CMBM", "CMBT", "CMC", "CMCL", "CMCM", "CMCO",
                     "CMCSA", "CMCT", "CME", "CMG", "CMI", "CMLS", "CMMB", "CMND", "CMP", "CMPO", "CMPOW", "CMPR",
                     "CMPS", "CMPX", "CMRE", "CMS", "CMSA", "CMSC", "CMSD", "CMT", "CMTG", "CMTL", "CMU", "CNA", "CNC",
                     "CNCK", "CNCKW", "CNDT", "CNET", "CNEY", "CNF", "CNFR", "CNFRZ", "CNH", "CNI", "CNK", "CNL", "CNM",
                     "CNMD", "CNNE", "CNO", "CNOB", "CNOBP", "CNP", "CNQ", "CNR", "CNS", "CNSP", "CNTA", "CNTB", "CNTM",
                     "CNTX", "CNTY", "CNVS", "CNX", "CNXC", "CNXN", "COCH", "COCHW", "COCO", "COCP", "CODA", "CODI",
                     "CODX", "COE", "COEP", "COEPW", "COF", "COFS", "COGT", "COHN", "COHR", "COHU", "COIN", "COKE",
                     "COLA", "COLAU", "COLB", "COLD", "COLL", "COLM", "COMM", "COMP", "CON", "COO", "COOK", "COOP",
                     "COOT", "COP", "COR", "CORT", "CORZ", "CORZW", "CORZZ", "COSM", "COST", "COTY", "COUR", "COYA",
                     "CP", "CPA", "CPAC", "CPAY", "CPB", "CPBI", "CPF", "CPHC", "CPHI", "CPIX", "CPK", "CPNG", "CPOP",
                     "CPRI", "CPRT", "CPRX", "CPS", "CPSH", "CPSS", "CPT", "CPZ", "CQP", "CR", "CRAI", "CRBD", "CRBG",
                     "CRBP", "CRBU", "CRC", "CRCT", "CRD", "CRDF", "CRDL", "CRDO", "CREG", "CRESW", "CRESY", "CREV",
                     "CREVW", "CREX", "CRF", "CRGO", "CRGOW", "CRGX", "CRGY", "CRH", "CRI", "CRIS", "CRK", "CRL", "CRM",
                     "CRMD", "CRML", "CRMLW", "CRMT", "CRNC", "CRNT", "CRNX", "CRON", "CROX", "CRS", "CRSP", "CRSR",
                     "CRT", "CRTO", "CRUS", "CRVL", "CRVO", "CRVS", "CRWD", "CRWS", "CRWV", "CSAI", "CSAN", "CSBR",
                     "CSCI", "CSCO", "CSGP", "CSGS", "CSIQ", "CSL", "CSPI", "CSQ", "CSR", "CSTE", "CSTL", "CSTM", "CSV",
                     "CSWC", "CSWCZ", "CSWI", "CSX", "CTA", "CTAS", "CTBB", "CTBI", "CTDD", "CTEV", "CTGO", "CTKB",
                     "CTLP", "CTM", "CTMX", "CTNM", "CTNT", "CTO", "CTOR", "CTOS", "CTRA", "CTRE", "CTRI", "CTRM",
                     "CTRN", "CTS", "CTSH", "CTSO", "CTVA", "CTXR", "CUB", "CUBA", "CUBB", "CUBE", "CUBI", "CUBWU",
                     "CUBWW", "CUE", "CUK", "CULP", "CUPR", "CURB", "CURI", "CURIW", "CURR", "CURV", "CUZ", "CVAC",
                     "CVBF", "CVCO", "CVE", "CVEO", "CVGI", "CVGW", "CVI", "CVKD", "CVLG", "CVLT", "CVM", "CVNA", "CVR",
                     "CVRX", "CVS", "CVU", "CVV", "CVX", "CW", "CWAN", "CWBC", "CWCO", "CWD", "CWEN", "CWH", "CWK",
                     "CWST", "CWT", "CX", "CXAI", "CXAIW", "CXDO", "CXE", "CXH", "CXM", "CXT", "CXW", "CYBN", "CYBR",
                     "CYCC", "CYCCP", "CYCN", "CYCU", "CYCUW", "CYD", "CYH", "CYN", "CYRX", "CYTK", "CZFS", "CZNC",
                     "CZR", "CZWI", "D", "DAC", "DADA", "DAIO", "DAKT", "DAL", "DALN", "DAN", "DAO", "DAR", "DARE",
                     "DASH", "DATS", "DATSW", "DAVA", "DAVE", "DAVEW", "DAWN", "DAY", "DB", "DBD", "DBI", "DBL", "DBRG",
                     "DBVT", "DBX", "DC", "DCBO", "DCGO", "DCI", "DCO", "DCOM", "DCOMG", "DCOMP", "DCTH", "DD", "DDC",
                     "DDD", "DDI", "DDL", "DDOG", "DDS", "DDT", "DE", "DEA", "DEC", "DECK", "DEI", "DELL", "DENN",
                     "DEO", "DERM", "DESP", "DEVS", "DFH", "DFIN", "DFLI", "DFLIW", "DFP", "DFS", "DG", "DGICA",
                     "DGICB", "DGII", "DGLY", "DGNX", "DGX", "DGXX", "DH", "DHAI", "DHAIW", "DHC", "DHCNI", "DHCNL",
                     "DHF", "DHI", "DHIL", "DHR", "DHT", "DHX", "DHY", "DIAX", "DIBS", "DIN", "DINO", "DIOD", "DIS",
                     "DIST", "DISTR", "DISTW", "DIT", "DJCO", "DJT", "DJTWW", "DK", "DKL", "DKNG", "DKS", "DLB", "DLHC",
                     "DLNG", "DLO", "DLPN", "DLR", "DLTH", "DLTR", "DLX", "DLY", "DMA", "DMAA", "DMAAR", "DMAAU",
                     "DMAC", "DMB", "DMF", "DMLP", "DMN", "DMO", "DMRC", "DMYY", "DNA", "DNB", "DNLI", "DNN", "DNOW",
                     "DNP", "DNTH", "DNUT", "DOC", "DOCN", "DOCS", "DOCU", "DOGZ", "DOLE", "DOMH", "DOMO", "DOOO",
                     "DORM", "DOUG", "DOV", "DOW", "DOX", "DOYU", "DPG", "DPRO", "DPZ", "DQ", "DRCT", "DRD", "DRDB",
                     "DRDBU", "DRH", "DRI", "DRIO", "DRMA", "DRMAW", "DRRX", "DRS", "DRTS", "DRUG", "DRVN", "DSGN",
                     "DSGR", "DSGX", "DSL", "DSM", "DSP", "DSS", "DSU", "DSWL", "DSX", "DSY", "DSYWW", "DT", "DTB",
                     "DTCK", "DTE", "DTF", "DTG", "DTI", "DTIL", "DTM", "DTSQ", "DTSS", "DTST", "DTW", "DUK", "DUKB",
                     "DUO", "DUOL", "DUOT", "DV", "DVA", "DVAX", "DVLT", "DVN", "DVS", "DWSN", "DWTX", "DX", "DXC",
                     "DXCM", "DXF", "DXLG", "DXPE", "DXR", "DXST", "DXYZ", "DY", "DYAI", "DYCQ", "DYN", "DYNX", "DYNXU",
                     "DYNXW", "E", "EA", "EAD", "EAF", "EAI", "EARN", "EAT", "EB", "EBAY", "EBC", "EBF", "EBMT", "EBON",
                     "EBR", "EBS", "EBTC", "EC", "ECAT", "ECBK", "ECC", "ECCC", "ECCF", "ECCU", "ECCV", "ECCW", "ECCX",
                     "ECDA", "ECDAW", "ECF", "ECG", "ECL", "ECO", "ECOR", "ECPG", "ECVT", "ECX", "ECXWW", "ED", "EDAP",
                     "EDBL", "EDBLW", "EDD", "EDF", "EDHL", "EDIT", "EDN", "EDRY", "EDSA", "EDTK", "EDU", "EDUC", "EE",
                     "EEA", "EEFT", "EEIQ", "EEX", "EFC", "EFOI", "EFR", "EFSC", "EFSCP", "EFSI", "EFT", "EFX", "EFXT",
                     "EG", "EGAN", "EGBN", "EGHT", "EGO", "EGP", "EGY", "EH", "EHAB", "EHC", "EHGO", "EHI", "EHLD",
                     "EHTH", "EIC", "EICA", "EICB", "EICC", "EIG", "EIIA", "EIM", "EIX", "EJH", "EKSO", "EL", "ELA",
                     "ELAB", "ELAN", "ELBM", "ELC", "ELDN", "ELEV", "ELF", "ELLO", "ELMD", "ELME", "ELP", "ELPC",
                     "ELPW", "ELS", "ELSE", "ELTK", "ELTX", "ELUT", "ELV", "ELVA", "ELVN", "ELWS", "EM", "EMBC", "EMCG",
                     "EMCGR", "EMD", "EME", "EMF", "EML", "EMN", "EMO", "EMP", "EMR", "EMX", "ENB", "ENFY", "ENGN",
                     "ENGNW", "ENGS", "ENIC", "ENJ", "ENLT", "ENLV", "ENO", "ENOV", "ENPH", "ENR", "ENS", "ENSC",
                     "ENSG", "ENTA", "ENTG", "ENTO", "ENTX", "ENVA", "ENVB", "ENVX", "ENX", "EOD", "EOG", "EOI", "EOLS",
                     "EONR", "EOS", "EOSE", "EOSEW", "EOT", "EP", "EPAC", "EPAM", "EPC", "EPD", "EPIX", "EPM", "EPOW",
                     "EPR", "EPRT", "EPRX", "EPSM", "EPSN", "EPWK", "EQ", "EQBK", "EQH", "EQIX", "EQNR", "EQR", "EQS",
                     "EQT", "EQV", "EQX", "ERAS", "ERC", "ERH", "ERIC", "ERIE", "ERII", "ERJ", "ERNA", "ERO", "ES",
                     "ESAB", "ESCA", "ESE", "ESEA", "ESGL", "ESGR", "ESGRO", "ESGRP", "ESHA", "ESI", "ESLA", "ESLT",
                     "ESNT", "ESOA", "ESP", "ESPR", "ESQ", "ESRT", "ESS", "ESSA", "ESTA", "ESTC", "ET", "ETB", "ETD",
                     "ETG", "ETI", "ETJ", "ETN", "ETNB", "ETO", "ETON", "ETR", "ETSY", "ETV", "ETW", "ETWO", "ETX",
                     "ETY", "EU", "EUDA", "EUDAW", "EURK", "EURKU", "EVAX", "EVBN", "EVC", "EVCM", "EVER", "EVEX",
                     "EVF", "EVG", "EVGN", "EVGO", "EVGOW", "EVH", "EVI", "EVLV", "EVLVW", "EVM", "EVN", "EVO", "EVOK",
                     "EVR", "EVRG", "EVRI", "EVT", "EVTC", "EVTL", "EVTV", "EVV", "EW", "EWBC", "EWCZ", "EWTX", "EXAS",
                     "EXC", "EXE", "EXEEL", "EXEL", "EXFY", "EXG", "EXK", "EXLS", "EXOD", "EXOZ", "EXP", "EXPD", "EXPE",
                     "EXPI", "EXPO", "EXR", "EXTR", "EYE", "EYEN", "EYPT", "EZGO", "EZPW", "F", "FA", "FAAS", "FAASW",
                     "FACT", "FAF", "FAMI", "FANG", "FARM", "FARO", "FAST", "FAT", "FATBB", "FATBP", "FATBW", "FATE",
                     "FATN", "FAX", "FBGL", "FBIN", "FBIO", "FBIOP", "FBIZ", "FBK", "FBLA", "FBLG", "FBNC", "FBP",
                     "FBRT", "FBRX", "FBYD", "FBYDW", "FC", "FCAP", "FCBC", "FCCO", "FCEL", "FCF", "FCFS", "FCN",
                     "FCNCA", "FCNCO", "FCNCP", "FCO", "FCPT", "FCRX", "FCT", "FCUV", "FCX", "FDBC", "FDMT", "FDP",
                     "FDSB", "FDUS", "FDX", "FE", "FEAM", "FEBO", "FEDU", "FEIM", "FELE", "FEMY", "FENC", "FENG", "FER",
                     "FERA", "FERAR", "FERAU", "FERG", "FET", "FF", "FFA", "FFAI", "FFAIW", "FFBC", "FFC", "FFIC",
                     "FFIN", "FFIV", "FFWM", "FG", "FGB", "FGBI", "FGBIP", "FGEN", "FGF", "FGI", "FGIWW", "FGL", "FGMC",
                     "FGMCR", "FGMCU", "FGN", "FGSN", "FHB", "FHI", "FHN", "FHTX", "FI", "FIBK", "FICO", "FIGS", "FIHL",
                     "FINS", "FINV", "FINW", "FIP", "FIS", "FISI", "FITB", "FITBI", "FITBO", "FITBP", "FIVE", "FIVN",
                     "FIX", "FIZZ", "FKWL", "FL", "FLC", "FLD", "FLDDW", "FLEX", "FLG", "FLGC", "FLGT", "FLIC", "FLL",
                     "FLNC", "FLNG", "FLNT", "FLO", "FLOC", "FLR", "FLS", "FLUT", "FLUX", "FLWS", "FLX", "FLXS", "FLYE",
                     "FLYW", "FLYX", "FMAO", "FMBH", "FMC", "FMN", "FMNB", "FMS", "FMST", "FMTO", "FMX", "FMY", "FN",
                     "FNB", "FND", "FNF", "FNGR", "FNKO", "FNLC", "FNV", "FNWB", "FNWD", "FOA", "FOF", "FOLD", "FONR",
                     "FOR", "FORA", "FORD", "FORL", "FORM", "FORR", "FORTY", "FOSL", "FOSLL", "FOUR", "FOX", "FOXA",
                     "FOXF", "FOXX", "FPAY", "FPF", "FPH", "FPI", "FR", "FRA", "FRAF", "FRBA", "FRD", "FRGE", "FRGT",
                     "FRHC", "FRME", "FRMEP", "FRO", "FROG", "FRPH", "FRPT", "FRSH", "FRST", "FRSX", "FRT", "FSBC",
                     "FSBW", "FSCO", "FSEA", "FSFG", "FSHPR", "FSHPU", "FSI", "FSK", "FSLR", "FSLY", "FSM", "FSP",
                     "FSS", "FSTR", "FSUN", "FSV", "FT", "FTAI", "FTAIM", "FTAIN", "FTCI", "FTDR", "FTEK", "FTEL",
                     "FTF", "FTFT", "FTHM", "FTHY", "FTI", "FTK", "FTLF", "FTNT", "FTRE", "FTS", "FTV", "FUBO", "FUFU",
                     "FUFUW", "FUL", "FULC", "FULT", "FULTP", "FUN", "FUNC", "FUND", "FURY", "FUSB", "FUTU", "FVCB",
                     "FVN", "FVNNR", "FVR", "FVRR", "FWONA", "FWONK", "FWRD", "FWRG", "FXNC", "FYBR", "G", "GAB",
                     "GABC", "GAIA", "GAIN", "GAINI", "GAINL", "GAINN", "GAINZ", "GALT", "GAM", "GAMB", "GAME", "GAN",
                     "GANX", "GAP", "GASS", "GATX", "GAU", "GAUZ", "GB", "GBAB", "GBCI", "GBDC", "GBIO", "GBLI", "GBR",
                     "GBTG", "GBX", "GCBC", "GCI", "GCL", "GCLWW", "GCMG", "GCMGW", "GCO", "GCT", "GCTK", "GCTS", "GCV",
                     "GD", "GDC", "GDDY", "GDEN", "GDEV", "GDHG", "GDL", "GDO", "GDOT", "GDRX", "GDS", "GDTC", "GDV",
                     "GDYN", "GE", "GECC", "GECCH", "GECCI", "GECCO", "GECCZ", "GEF", "GEG", "GEGGL", "GEHC", "GEL",
                     "GELS", "GEN", "GENC", "GENI", "GENK", "GENVR", "GEO", "GEOS", "GERN", "GES", "GETY", "GEV",
                     "GEVO", "GF", "GFAI", "GFAIW", "GFF", "GFI", "GFL", "GFR", "GFS", "GGAL", "GGB", "GGG", "GGN",
                     "GGR", "GGROW", "GGT", "GGZ", "GH", "GHC", "GHG", "GHI", "GHLD", "GHM", "GHRS", "GHY", "GIB",
                     "GIC", "GIFI", "GIFT", "GIG", "GIGGU", "GIGM", "GIII", "GIL", "GILD", "GILT", "GIPR", "GIS",
                     "GITS", "GJP", "GJR", "GJS", "GKOS", "GL", "GLAD", "GLADZ", "GLBE", "GLBS", "GLBZ", "GLDD", "GLDG",
                     "GLE", "GLMD", "GLNG", "GLO", "GLOB", "GLOP", "GLP", "GLPG", "GLPI", "GLQ", "GLRE", "GLSI", "GLTO",
                     "GLU", "GLUE", "GLV", "GLW", "GLXG", "GLYC", "GM", "GMAB", "GME", "GMED", "GMGI", "GMHS", "GMM",
                     "GMRE", "GMS", "GNE", "GNFT", "GNK", "GNL", "GNLN", "GNLX", "GNPX", "GNRC", "GNS", "GNSS", "GNT",
                     "GNTA", "GNTX", "GNTY", "GNW", "GO", "GOCO", "GOF", "GOGL", "GOGO", "GOLD", "GOLF", "GOOD",
                     "GOODN", "GOODO", "GOOG", "GOOGL", "GOOS", "GORO", "GORV", "GOSS", "GOTU", "GOVX", "GOVXW", "GP",
                     "GPAT", "GPATW", "GPC", "GPCR", "GPI", "GPJA", "GPK", "GPMT", "GPN", "GPOR", "GPRE", "GPRK",
                     "GPRO", "GPUS", "GRAB", "GRABW", "GRAF", "GRAL", "GRBK", "GRC", "GRCE", "GRDN", "GREE", "GREEL",
                     "GRF", "GRFS", "GRFX", "GRI", "GRMN", "GRND", "GRNQ", "GRNT", "GRO", "GROV", "GROW", "GROY",
                     "GRPN", "GRRR", "GRRRW", "GRVY", "GRWG", "GRX", "GRYP", "GS", "GSAT", "GSBC", "GSBD", "GSHD",
                     "GSHRU", "GSIT", "GSIW", "GSK", "GSL", "GSM", "GSRT", "GSRTR", "GSRTU", "GSUN", "GT", "GTBP",
                     "GTE", "GTEC", "GTES", "GTI", "GTIM", "GTLB", "GTLS", "GTN", "GTX", "GTY", "GUG", "GURE", "GUT",
                     "GUTS", "GV", "GVA", "GVH", "GWAV", "GWH", "GWRE", "GWRS", "GWW", "GXAI", "GXO", "GYRE", "GYRO",
                     "H", "HAE", "HAFC", "HAFN", "HAIN", "HAL", "HALO", "HAO", "HAS", "HASI", "HAYW", "HBAN", "HBANL",
                     "HBANM", "HBANP", "HBB", "HBCP", "HBI", "HBIO", "HBM", "HBNC", "HBT", "HCA", "HCAI", "HCAT", "HCC",
                     "HCI", "HCKT", "HCM", "HCSG", "HCTI", "HCWB", "HCWC", "HCXY", "HD", "HDB", "HDL", "HDSN", "HE",
                     "HEES", "HEI", "HELE", "HEPA", "HEPS", "HEQ", "HES", "HESM", "HFBL", "HFFG", "HFRO", "HFWA", "HG",
                     "HGBL", "HGLB", "HGTY", "HGV", "HHH", "HHS", "HI", "HIFS", "HIG", "HIHO", "HII", "HIMS", "HIMX",
                     "HIO", "HIPO", "HIT", "HITI", "HIVE", "HIW", "HIX", "HKD", "HKIT", "HKPD", "HL", "HLF", "HLI",
                     "HLIO", "HLIT", "HLLY", "HLMN", "HLN", "HLNE", "HLP", "HLT", "HLVX", "HLX", "HLXB", "HMC", "HMN",
                     "HMR", "HMST", "HMY", "HNI", "HNNA", "HNNAZ", "HNRG", "HNST", "HNVR", "HNW", "HOFT", "HOFV",
                     "HOFVW", "HOG", "HOLO", "HOLOW", "HOLX", "HOMB", "HON", "HOND", "HONDU", "HONDW", "HONE", "HOOD",
                     "HOOK", "HOPE", "HOTH", "HOUR", "HOUS", "HOV", "HOVNP", "HOVR", "HOVRW", "HOWL", "HP", "HPAI",
                     "HPE", "HPF", "HPH", "HPI", "HPK", "HPKEW", "HPP", "HPQ", "HPS", "HQH", "HQI", "HQL", "HQY", "HR",
                     "HRB", "HRI", "HRL", "HRMY", "HROW", "HROWL", "HROWM", "HRTG", "HRTX", "HRZN", "HSAI", "HSBC",
                     "HSCS", "HSDT", "HSHP", "HSIC", "HSII", "HSON", "HSPO", "HSPOR", "HST", "HSTM", "HSY", "HTB",
                     "HTBK", "HTCO", "HTCR", "HTD", "HTFC", "HTGC", "HTH", "HTHT", "HTLD", "HTLM", "HTOO", "HTZ",
                     "HTZWW", "HUBB", "HUBC", "HUBCW", "HUBG", "HUBS", "HUDI", "HUHU", "HUIZ", "HUM", "HUMA", "HUMAW",
                     "HUN", "HURA", "HURC", "HURN", "HUSA", "HUT", "HUYA", "HVII", "HVIIR", "HVIIU", "HVT", "HWBK",
                     "HWC", "HWCPZ", "HWH", "HWKN", "HWM", "HXHX", "HXL", "HY", "HYAC", "HYFM", "HYI", "HYLN", "HYMC",
                     "HYMCL", "HYMCW", "HYPR", "HYT", "HZO", "I", "IAC", "IAE", "IAF", "IAG", "IART", "IAS", "IAUX",
                     "IBAC", "IBACR", "IBCP", "IBEX", "IBG", "IBIO", "IBKR", "IBM", "IBN", "IBO", "IBOC", "IBP", "IBRX",
                     "IBTA", "ICAD", "ICCC", "ICCM", "ICCT", "ICE", "ICFI", "ICG", "ICHR", "ICL", "ICLR", "ICMB",
                     "ICON", "ICR", "ICU", "ICUCW", "ICUI", "IDA", "IDAI", "IDCC", "IDE", "IDN", "IDR", "IDT", "IDXX",
                     "IDYA", "IE", "IEP", "IESC", "IEX", "IFBD", "IFF", "IFN", "IFRX", "IFS", "IGA", "IGC", "IGD",
                     "IGI", "IGIC", "IGMS", "IGR", "IGT", "IH", "IHD", "IHG", "IHRT", "IHS", "IHT", "IIF", "III",
                     "IIIN", "IIIV", "IIM", "IINN", "IIPR", "IKNA", "IKT", "ILAG", "ILLR", "ILLRW", "ILMN", "ILPT",
                     "IMAB", "IMAX", "IMCC", "IMCR", "IMG", "IMKTA", "IMMP", "IMMR", "IMMX", "IMNM", "IMNN", "IMO",
                     "IMOS", "IMPP", "IMPPP", "IMRN", "IMRX", "IMTE", "IMTX", "IMUX", "IMVT", "IMXI", "INAB", "INBK",
                     "INBKZ", "INBS", "INBX", "INCR", "INCY", "INDB", "INDI", "INDO", "INDP", "INDV", "INFA", "INFU",
                     "INFY", "ING", "INGM", "INGN", "INGR", "INHD", "INKT", "INLF", "INLX", "INM", "INMB", "INMD",
                     "INN", "INNV", "INO", "INOD", "INR", "INSE", "INSG", "INSM", "INSP", "INSW", "INTA", "INTC",
                     "INTG", "INTJ", "INTR", "INTS", "INTT", "INTU", "INTZ", "INUV", "INV", "INVA", "INVE", "INVH",
                     "INVX", "INVZ", "INVZW", "INZY", "IOBT", "IONQ", "IONR", "IONS", "IOR", "IOSP", "IOT", "IOTR",
                     "IOVA", "IP", "IPA", "IPAR", "IPDN", "IPG", "IPGP", "IPHA", "IPI", "IPM", "IPSC", "IPW", "IPWR",
                     "IPX", "IQ", "IQI", "IQV", "IR", "IRBT", "IRD", "IRDM", "IREN", "IRIX", "IRM", "IRMD", "IROH",
                     "IROHR", "IROHW", "IRON", "IROQ", "IRS", "IRT", "IRTC", "IRWD", "ISD", "ISPC", "ISPO", "ISPR",
                     "ISRG", "ISRL", "ISRLW", "ISSC", "ISTR", "IT", "ITGR", "ITIC", "ITOS", "ITP", "ITRG", "ITRI",
                     "ITRM", "ITRN", "ITT", "ITUB", "ITW", "IVA", "IVCA", "IVDA", "IVP", "IVR", "IVT", "IVVD", "IVZ",
                     "IX", "IXHL", "IZEA", "IZM", "J", "JACK", "JACS", "JAGX", "JAKK", "JAMF", "JANX", "JAZZ", "JBDI",
                     "JBGS", "JBHT", "JBI", "JBL", "JBLU", "JBSS", "JBTM", "JCE", "JCI", "JCSE", "JCTC", "JD", "JDZG",
                     "JEF", "JELD", "JEQ", "JFB", "JFBR", "JFBRW", "JFIN", "JFR", "JFU", "JG", "JGH", "JHG", "JHI",
                     "JHS", "JHX", "JILL", "JJSF", "JKHY", "JKS", "JL", "JLL", "JLS", "JMIA", "JMM", "JMSB", "JNJ",
                     "JNPR", "JNVR", "JOB", "JOBY", "JOE", "JOF", "JOUT", "JOYY", "JPC", "JPI", "JPM", "JQC", "JRI",
                     "JRS", "JRSH", "JRVR", "JSM", "JSPR", "JTAI", "JUNE", "JUNS", "JVA", "JWEL", "JWN", "JXG", "JXN",
                     "JYD", "JYNT", "JZ", "JZXN", "K", "KAI", "KALA", "KALU", "KALV", "KAPA", "KAR", "KARO", "KAVL",
                     "KB", "KBDC", "KBH", "KBR", "KBSX", "KC", "KD", "KDLY", "KDLYW", "KDP", "KE", "KELYA", "KELYB",
                     "KEN", "KEP", "KEQU", "KEX", "KEY", "KEYS", "KF", "KFFB", "KFII", "KFIIR", "KFIIU", "KFRC", "KFS",
                     "KFY", "KGC", "KGEI", "KGS", "KHC", "KIDS", "KIDZ", "KIDZW", "KIM", "KIND", "KINS", "KIO", "KIRK",
                     "KITT", "KITTW", "KKR", "KKRS", "KLAC", "KLC", "KLG", "KLIC", "KLRS", "KLTO", "KLTOW", "KLTR",
                     "KLXE", "KMB", "KMDA", "KMI", "KMPB", "KMPR", "KMT", "KMTS", "KMX", "KN", "KNDI", "KNF", "KNOP",
                     "KNSA", "KNSL", "KNTK", "KNW", "KNX", "KO", "KOD", "KODK", "KOF", "KOP", "KOPN", "KORE", "KOS",
                     "KOSS", "KPLT", "KPLTW", "KPRX", "KPTI", "KR", "KRC", "KREF", "KRG", "KRKR", "KRMD", "KRMN",
                     "KRNT", "KRNY", "KRO", "KRON", "KROS", "KRP", "KRRO", "KRT", "KRUS", "KRYS", "KSCP", "KSPI", "KSS",
                     "KT", "KTB", "KTCC", "KTF", "KTH", "KTN", "KTOS", "KTTA", "KTTAW", "KUKE", "KULR", "KURA", "KVACW",
                     "KVHI", "KVUE", "KVYO", "KW", "KWE", "KWESW", "KWR", "KXIN", "KYMR", "KYN", "KYTX", "KZIA", "KZR",
                     "L", "LAB", "LAC", "LAD", "LADR", "LAES", "LAKE", "LAMR", "LANC", "LAND", "LANDM", "LANDO",
                     "LANDP", "LANV", "LAR", "LARK", "LASE", "LASR", "LAUR", "LAW", "LAZ", "LAZR", "LB", "LBGJ",
                     "LBRDA", "LBRDK", "LBRDP", "LBRT", "LBTYA", "LBTYB", "LBTYK", "LC", "LCFY", "LCFYW", "LCID",
                     "LCII", "LCNB", "LCTX", "LCUT", "LDI", "LDOS", "LDP", "LDTC", "LDTCW", "LDWY", "LE", "LEA", "LECO",
                     "LEDS", "LEE", "LEG", "LEGH", "LEGN", "LEGT", "LEN", "LENZ", "LEO", "LESL", "LEU", "LEVI", "LEXX",
                     "LEXXW", "LFCR", "LFMD", "LFMDP", "LFST", "LFT", "LFUS", "LFVN", "LFWD", "LGCB", "LGCL", "LGCY",
                     "LGHL", "LGHLW", "LGI", "LGIH", "LGL", "LGMK", "LGND", "LGO", "LGPS", "LGVN", "LH", "LHSW", "LHX",
                     "LI", "LICN", "LIDR", "LIDRW", "LIEN", "LIF", "LII", "LILA", "LILAK", "LIN", "LINC", "LIND",
                     "LINE", "LINK", "LION", "LIPO", "LIQT", "LITB", "LITE", "LITM", "LIVE", "LIVN", "LIXT", "LKFN",
                     "LKQ", "LLY", "LLYVA", "LLYVK", "LMAT", "LMB", "LMFA", "LMND", "LMNR", "LMT", "LNC", "LND", "LNG",
                     "LNKB", "LNKS", "LNN", "LNSR", "LNT", "LNTH", "LNW", "LNZA", "LNZAW", "LOAN", "LOAR", "LOB",
                     "LOBO", "LOCL", "LOCO", "LODE", "LOGC", "LOGI", "LOKV", "LOKVU", "LOMA", "LOOP", "LOPE", "LOT",
                     "LOVE", "LOW", "LPA", "LPAA", "LPAAU", "LPAAW", "LPBB", "LPBBU", "LPBBW", "LPCN", "LPG", "LPL",
                     "LPLA", "LPRO", "LPSN", "LPTH", "LPTX", "LPX", "LQDA", "LQDT", "LRCX", "LRE", "LRFC", "LRHC",
                     "LRMR", "LRN", "LSAK", "LSB", "LSBK", "LSBPW", "LSCC", "LSE", "LSEA", "LSEAW", "LSF", "LSH",
                     "LSPD", "LSTA", "LSTR", "LTBR", "LTC", "LTH", "LTM", "LTRN", "LTRX", "LTRY", "LTRYW", "LU", "LUCD",
                     "LUCK", "LUCY", "LUCYW", "LUD", "LULU", "LUMN", "LUNG", "LUNR", "LUV", "LVLU", "LVO", "LVRO",
                     "LVROW", "LVS", "LVTX", "LVWR", "LW", "LWAY", "LWLG", "LX", "LXEH", "LXEO", "LXFR", "LXP", "LXRX",
                     "LXU", "LYB", "LYEL", "LYFT", "LYG", "LYRA", "LYTS", "LYV", "LZ", "LZB", "LZM", "LZMH", "M", "MA",
                     "MAA", "MAC", "MAG", "MAGN", "MAIA", "MAIN", "MAMA", "MAMO", "MAN", "MANH", "MANU", "MAPS",
                     "MAPSW", "MAR", "MARA", "MARPS", "MAS", "MASI", "MASK", "MASS", "MAT", "MATH", "MATV", "MATW",
                     "MATX", "MAV", "MAX", "MAXN", "MAYA", "MAYAU", "MAYS", "MAZE", "MB", "MBAV", "MBAVU", "MBAVW",
                     "MBBC", "MBC", "MBCN", "MBI", "MBIN", "MBINL", "MBINM", "MBINN", "MBIO", "MBLY", "MBNKP", "MBOT",
                     "MBRX", "MBUU", "MBWM", "MBX", "MC", "MCB", "MCBS", "MCD", "MCFT", "MCHP", "MCHPP", "MCHX", "MCI",
                     "MCK", "MCN", "MCO", "MCR", "MCRB", "MCRI", "MCRP", "MCS", "MCTR", "MCVT", "MCW", "MCY", "MD",
                     "MDAI", "MDAIW", "MDB", "MDBH", "MDCX", "MDCXW", "MDGL", "MDIA", "MDLZ", "MDRR", "MDT", "MDU",
                     "MDV", "MDWD", "MDXG", "MDXH", "MEC", "MED", "MEDP", "MEG", "MEGI", "MEGL", "MEI", "MEIP", "MELI",
                     "MEOH", "MER", "MERC", "MESA", "MESO", "MET", "META", "METC", "METCB", "METCL", "METCZ", "MFA",
                     "MFAN", "MFAO", "MFC", "MFG", "MFH", "MFI", "MFIC", "MFICL", "MFIN", "MFM", "MG", "MGA", "MGEE",
                     "MGF", "MGIC", "MGIH", "MGLD", "MGM", "MGNI", "MGNX", "MGPI", "MGR", "MGRB", "MGRC", "MGRD",
                     "MGRE", "MGRM", "MGRX", "MGTX", "MGX", "MGY", "MGYR", "MHD", "MHF", "MHH", "MHI", "MHK", "MHLA",
                     "MHLD", "MHN", "MHNC", "MHO", "MHUA", "MI", "MIDD", "MIGI", "MIMI", "MIN", "MIND", "MIO", "MIR",
                     "MIRA", "MIRM", "MIST", "MITK", "MITN", "MITP", "MITQ", "MITT", "MIY", "MKC", "MKDW", "MKDWW",
                     "MKL", "MKSI", "MKTW", "MKTX", "MKZR", "MLAB", "MLAC", "MLACR", "MLACU", "MLCO", "MLEC", "MLGO",
                     "MLI", "MLKN", "MLM", "MLNK", "MLP", "MLR", "MLSS", "MLTX", "MLYS", "MMA", "MMC", "MMD", "MMI",
                     "MMLP", "MMM", "MMS", "MMSI", "MMT", "MMU", "MMYT", "MNDO", "MNDR", "MNDY", "MNKD", "MNMD", "MNOV",
                     "MNPR", "MNR", "MNRO", "MNSB", "MNSO", "MNST", "MNTK", "MNTS", "MNTSW", "MNY", "MNYWW", "MO",
                     "MOB", "MOBBW", "MOBX", "MOD", "MODD", "MODG", "MODV", "MOFG", "MOGO", "MOGU", "MOH", "MOLN",
                     "MOMO", "MORN", "MOS", "MOV", "MOVE", "MP", "MPA", "MPAA", "MPB", "MPC", "MPLX", "MPTI", "MPU",
                     "MPV", "MPW", "MPWR", "MPX", "MQ", "MQT", "MQY", "MRAM", "MRBK", "MRC", "MRCC", "MRCY", "MREO",
                     "MRIN", "MRK", "MRKR", "MRM", "MRNA", "MRNO", "MRNOW", "MRP", "MRSN", "MRT", "MRTN", "MRUS",
                     "MRVI", "MRVL", "MRX", "MS", "MSA", "MSAI", "MSB", "MSBI", "MSBIP", "MSC", "MSCI", "MSD", "MSDL",
                     "MSEX", "MSFT", "MSGE", "MSGM", "MSGS", "MSI", "MSIF", "MSM", "MSN", "MSPR", "MSPRW", "MSPRZ",
                     "MSS", "MSTR", "MSW", "MT", "MTA", "MTAL", "MTB", "MTC", "MTCH", "MTD", "MTDR", "MTEK", "MTEKW",
                     "MTEN", "MTEX", "MTG", "MTH", "MTLS", "MTN", "MTNB", "MTR", "MTRN", "MTRX", "MTSI", "MTSR", "MTUS",
                     "MTVA", "MTW", "MTX", "MTZ", "MU", "MUA", "MUC", "MUE", "MUFG", "MUJ", "MULN", "MUR", "MURA",
                     "MUSA", "MUX", "MVBF", "MVF", "MVIS", "MVO", "MVST", "MVSTW", "MVT", "MWA", "MWG", "MWYN", "MX",
                     "MXC", "MXCT", "MXE", "MXF", "MXL", "MYD", "MYE", "MYFW", "MYGN", "MYI", "MYN", "MYND", "MYNZ",
                     "MYO", "MYPS", "MYPSW", "MYRG", "MYSZ", "MYTE", "N", "NA", "NAAS", "NABL", "NAC", "NAD", "NAGE",
                     "NAII", "NAK", "NAMI", "NAMS", "NAMSW", "NAN", "NAOV", "NAT", "NATH", "NATL", "NATR", "NAUT",
                     "NAVI", "NAYA", "NAZ", "NB", "NBB", "NBBK", "NBH", "NBHC", "NBIS", "NBIX", "NBN", "NBR", "NBTB",
                     "NBTX", "NBXG", "NBY", "NC", "NCA", "NCDL", "NCEW", "NCI", "NCIQ", "NCL", "NCLH", "NCMI", "NCNA",
                     "NCNO", "NCPL", "NCRA", "NCSM", "NCT", "NCTY", "NCV", "NCZ", "NDAQ", "NDLS", "NDMO", "NDRA",
                     "NDSN", "NE", "NEA", "NECB", "NEE", "NEGG", "NEHC", "NEHCW", "NEM", "NEN", "NEO", "NEOG", "NEON",
                     "NEOV", "NEOVW", "NEPH", "NERV", "NESR", "NESRW", "NET", "NETD", "NETDW", "NEU", "NEUE", "NEUP",
                     "NEWP", "NEWT", "NEWTG", "NEWTH", "NEWTI", "NEWTZ", "NEXA", "NEXN", "NEXT", "NFBK", "NFE", "NFG",
                     "NFGC", "NFJ", "NFLX", "NG", "NGD", "NGG", "NGL", "NGNE", "NGS", "NGVC", "NGVT", "NHC", "NHI",
                     "NHIC", "NHICU", "NHICW", "NHPAP", "NHPBP", "NHS", "NHTC", "NI", "NIC", "NICE", "NIE", "NIM",
                     "NINE", "NIO", "NIOBW", "NIPG", "NISN", "NITO", "NIU", "NIVF", "NIXX", "NIXXW", "NJR", "NKE",
                     "NKSH", "NKTR", "NKTX", "NKX", "NL", "NLOP", "NLSP", "NLSPW", "NLY", "NMAI", "NMAX", "NMCO",
                     "NMFC", "NMFCZ", "NMG", "NMI", "NMIH", "NML", "NMM", "NMR", "NMRA", "NMRK", "NMS", "NMT", "NMTC",
                     "NMZ", "NN", "NNAVW", "NNBR", "NNDM", "NNE", "NNI", "NNN", "NNNN", "NNOX", "NNVC", "NNY", "NOA",
                     "NOAH", "NOC", "NODK", "NOEM", "NOEMR", "NOG", "NOK", "NOM", "NOMD", "NOTE", "NOTV", "NOV", "NOVA",
                     "NOVT", "NOW", "NPACU", "NPB", "NPCE", "NPCT", "NPFD", "NPK", "NPKI", "NPO", "NPV", "NPWR", "NQP",
                     "NRC", "NRDS", "NRDY", "NREF", "NRG", "NRGV", "NRIM", "NRIX", "NRK", "NRO", "NRP", "NRSN", "NRSNW",
                     "NRT", "NRUC", "NRXP", "NRXPW", "NRXS", "NSA", "NSC", "NSIT", "NSP", "NSPR", "NSSC", "NSTS",
                     "NSYS", "NTAP", "NTB", "NTCL", "NTCT", "NTES", "NTGR", "NTHI", "NTIC", "NTIP", "NTLA", "NTNX",
                     "NTR", "NTRA", "NTRB", "NTRP", "NTRS", "NTRSO", "NTST", "NTWK", "NTWO", "NTWOU", "NTWOW", "NTZ",
                     "NU", "NUE", "NUKK", "NUKKW", "NURO", "NUS", "NUTX", "NUV", "NUVB", "NUVL", "NUW", "NUWE", "NVA",
                     "NVAWW", "NVAX", "NVCR", "NVCT", "NVDA", "NVEC", "NVEE", "NVFY", "NVG", "NVGS", "NVMI", "NVNI",
                     "NVNIW", "NVNO", "NVO", "NVR", "NVRI", "NVS", "NVST", "NVT", "NVTS", "NVVE", "NVVEW", "NVX",
                     "NWBI", "NWE", "NWFL", "NWG", "NWGL", "NWL", "NWN", "NWPX", "NWS", "NWSA", "NWTG", "NWTN", "NWTNW",
                     "NX", "NXC", "NXDT", "NXE", "NXG", "NXGL", "NXGLW", "NXJ", "NXL", "NXN", "NXP", "NXPI", "NXPL",
                     "NXPLW", "NXRT", "NXST", "NXT", "NXTC", "NXTT", "NXXT", "NYAX", "NYC", "NYMT", "NYMTG", "NYMTI",
                     "NYMTL", "NYMTM", "NYMTN", "NYMTZ", "NYT", "NYXH", "NZF", "O", "OABI", "OACC", "OACCU", "OAK",
                     "OAKU", "OAKUR", "OAKUW", "OB", "OBDC", "OBE", "OBIO", "OBK", "OBLG", "OBT", "OC", "OCC", "OCCI",
                     "OCCIM", "OCCIN", "OCCIO", "OCFC", "OCFCP", "OCFT", "OCG", "OCGN", "OCS", "OCSL", "OCTO", "OCUL",
                     "OCX", "ODC", "ODD", "ODFL", "ODP", "ODV", "ODVWZ", "ODYS", "OEC", "OESX", "OFG", "OFIX", "OFLX",
                     "OFS", "OGE", "OGEN", "OGI", "OGN", "OGS", "OHI", "OI", "OIA", "OII", "OIS", "OKE", "OKLO", "OKTA",
                     "OKUR", "OKYO", "OLB", "OLED", "OLLI", "OLMA", "OLN", "OLO", "OLP", "OLPX", "OM", "OMAB", "OMC",
                     "OMCC", "OMCL", "OMER", "OMEX", "OMF", "OMH", "OMI", "ON", "ONB", "ONBPO", "ONBPP", "ONC", "ONCO",
                     "ONCY", "ONDS", "ONEG", "ONEW", "ONFO", "ONIT", "ONL", "ONMD", "ONON", "ONTF", "ONTO", "OOMA",
                     "OP", "OPAD", "OPAL", "OPBK", "OPCH", "OPEN", "OPFI", "OPHC", "OPI", "OPINL", "OPK", "OPOF", "OPP",
                     "OPRA", "OPRT", "OPRX", "OPTN", "OPTT", "OPTX", "OPTXW", "OPXS", "OPY", "OR", "ORA", "ORC", "ORCL",
                     "ORGN", "ORGNW", "ORGO", "ORI", "ORIC", "ORIS", "ORKA", "ORKT", "ORLA", "ORLY", "ORMP", "ORN",
                     "ORRF", "OS", "OSBC", "OSCR", "OSIS", "OSK", "OSPN", "OSRH", "OSRHW", "OSS", "OST", "OSTX", "OSUR",
                     "OSW", "OTEX", "OTIS", "OTLK", "OTLY", "OTRK", "OTTR", "OUST", "OUSTW", "OUSTZ", "OUT", "OVBC",
                     "OVID", "OVLY", "OVV", "OWL", "OWLT", "OXBR", "OXLC", "OXLCG", "OXLCI", "OXLCL", "OXLCN", "OXLCO",
                     "OXLCP", "OXLCZ", "OXM", "OXSQ", "OXSQG", "OXSQZ", "OXY", "OZ", "OZK", "OZKAP", "P", "PAA", "PAAS",
                     "PAC", "PACB", "PACK", "PACS", "PAG", "PAGP", "PAGS", "PAHC", "PAI", "PAL", "PALI", "PAM", "PAMT",
                     "PANL", "PANW", "PAPL", "PAR", "PARA", "PARAA", "PARR", "PASG", "PATH", "PATK", "PAVM", "PAVMZ",
                     "PAVS", "PAX", "PAXS", "PAY", "PAYC", "PAYO", "PAYS", "PAYX", "PB", "PBA", "PBBK", "PBF", "PBFS",
                     "PBH", "PBHC", "PBI", "PBM", "PBMWW", "PBPB", "PBR", "PBT", "PBYI", "PC", "PCAR", "PCB", "PCF",
                     "PCG", "PCH", "PCK", "PCLA", "PCM", "PCN", "PCOR", "PCQ", "PCRX", "PCSA", "PCSC", "PCT", "PCTTW",
                     "PCTY", "PCVX", "PCYO", "PD", "PDCC", "PDD", "PDEX", "PDFS", "PDI", "PDLB", "PDM", "PDO", "PDPA",
                     "PDS", "PDSB", "PDT", "PDX", "PDYN", "PDYNW", "PEB", "PEBK", "PEBO", "PECO", "PED", "PEG", "PEGA",
                     "PEN", "PENG", "PENN", "PEO", "PEP", "PEPG", "PERF", "PERI", "PESI", "PET", "PETS", "PETWW",
                     "PETZ", "PFAI", "PFBC", "PFD", "PFE", "PFG", "PFGC", "PFH", "PFIS", "PFL", "PFLT", "PFN", "PFO",
                     "PFS", "PFSI", "PFX", "PFXNZ", "PG", "PGC", "PGEN", "PGHL", "PGNY", "PGP", "PGR", "PGRE", "PGY",
                     "PGYWW", "PGZ", "PH", "PHAR", "PHAT", "PHD", "PHG", "PHGE", "PHH", "PHI", "PHIN", "PHIO", "PHK",
                     "PHLT", "PHM", "PHR", "PHT", "PHUN", "PHVS", "PHX", "PI", "PII", "PIII", "PIIIW", "PIM", "PINC",
                     "PINE", "PINS", "PIPR", "PITA", "PITAW", "PJT", "PK", "PKBK", "PKE", "PKG", "PKOH", "PKST", "PKX",
                     "PL", "PLAB", "PLAG", "PLAY", "PLBC", "PLBY", "PLCE", "PLD", "PLG", "PLL", "PLMK", "PLMKU",
                     "PLMKW", "PLMR", "PLNT", "PLOW", "PLPC", "PLRX", "PLRZ", "PLSE", "PLTK", "PLTR", "PLUG", "PLUR",
                     "PLUS", "PLUT", "PLX", "PLXS", "PLYA", "PLYM", "PM", "PMAX", "PMCB", "PMEC", "PMF", "PML", "PMM",
                     "PMN", "PMNT", "PMO", "PMT", "PMTS", "PMTU", "PMTV", "PMVP", "PMX", "PN", "PNBK", "PNC", "PNF",
                     "PNFP", "PNFPP", "PNI", "PNNT", "PNR", "PNRG", "PNTG", "PNW", "POAI", "POCI", "PODC", "PODD",
                     "POET", "POLA", "POLE", "POLEW", "PONY", "POOL", "POR", "POST", "POWI", "POWL", "POWW", "POWWP",
                     "PPBI", "PPBT", "PPC", "PPG", "PPIH", "PPL", "PPSI", "PPT", "PPTA", "PR", "PRA", "PRAA", "PRAX",
                     "PRCH", "PRCT", "PRDO", "PRE", "PRENW", "PRFX", "PRG", "PRGO", "PRGS", "PRH", "PRI", "PRIF",
                     "PRIM", "PRK", "PRKS", "PRLB", "PRLD", "PRM", "PRMB", "PRME", "PRO", "PROF", "PROK", "PROP",
                     "PROV", "PRPH", "PRPL", "PRPO", "PRQR", "PRS", "PRSO", "PRSU", "PRT", "PRTA", "PRTC", "PRTG",
                     "PRTH", "PRTS", "PRU", "PRVA", "PRZO", "PSA", "PSBD", "PSEC", "PSF", "PSFE", "PSHG", "PSIG",
                     "PSIX", "PSMT", "PSN", "PSNL", "PSNY", "PSNYW", "PSO", "PSQH", "PSTG", "PSTL", "PSTV", "PSX", "PT",
                     "PTA", "PTC", "PTCT", "PTEN", "PTGX", "PTHL", "PTIX", "PTIXW", "PTLE", "PTLO", "PTMN", "PTN",
                     "PTON", "PTPI", "PTY", "PUBM", "PUK", "PULM", "PUMP", "PVBC", "PVH", "PVL", "PVLA", "PW", "PWM",
                     "PWOD", "PWP", "PWR", "PX", "PXLW", "PXS", "PXSAW", "PYN", "PYPD", "PYPL", "PYXS", "PZC", "PZG",
                     "PZZA", "Q", "QBTS", "QCOM", "QCRH", "QD", "QDEL", "QETA", "QETAU", "QFIN", "QGEN", "QH", "QIPT",
                     "QLGN", "QLYS", "QMCO", "QMMM", "QNCX", "QNRX", "QNST", "QNTM", "QQQX", "QRHC", "QRVO", "QS",
                     "QSG", "QSI", "QSIAW", "QSR", "QTRX", "QTTB", "QTWO", "QUAD", "QUBT", "QUIK", "QURE", "QVCC",
                     "QVCD", "QVCGA", "QVCGB", "QVCGP", "QXO", "R", "RA", "RAC", "RACE", "RADX", "RAIL", "RAIN",
                     "RAINW", "RAMP", "RAND", "RANG", "RANGU", "RANI", "RAPP", "RAPT", "RARE", "RAVE", "RAY", "RAYA",
                     "RBA", "RBB", "RBBN", "RBC", "RBCAA", "RBKB", "RBLX", "RBNE", "RBOT", "RBRK", "RC", "RCAT", "RCB",
                     "RCC", "RCD", "RCEL", "RCG", "RCI", "RCKT", "RCKTW", "RCKY", "RCL", "RCMT", "RCON", "RCS", "RCT",
                     "RCUS", "RDAC", "RDCM", "RDDT", "RDFN", "RDGT", "RDHL", "RDI", "RDIB", "RDN", "RDNT", "RDUS",
                     "RDVT", "RDW", "RDWR", "RDY", "RDZN", "RDZNW", "REAL", "REAX", "REBN", "RECT", "REE", "REFI",
                     "REFR", "REG", "REGCO", "REGCP", "REGN", "REI", "REKR", "RELI", "RELL", "RELX", "RELY", "RENB",
                     "RENE", "RENEW", "RENT", "REPL", "REPX", "RERE", "RES", "RETO", "REVB", "REVBW", "REVG", "REX",
                     "REXR", "REYN", "REZI", "RF", "RFAI", "RFAIR", "RFI", "RFIL", "RFL", "RFM", "RFMZ", "RGA", "RGC",
                     "RGCO", "RGEN", "RGLD", "RGLS", "RGNX", "RGP", "RGR", "RGS", "RGT", "RGTI", "RGTIW", "RH", "RHI",
                     "RHLD", "RHP", "RICK", "RIG", "RIGL", "RILY", "RILYG", "RILYK", "RILYL", "RILYN", "RILYP", "RILYT",
                     "RILYZ", "RIME", "RIO", "RIOT", "RITM", "RITR", "RIV", "RIVN", "RJF", "RKDA", "RKLB", "RKT", "RL",
                     "RLAY", "RLGT", "RLI", "RLJ", "RLMD", "RLTY", "RLX", "RLYB", "RM", "RMAX", "RMBI", "RMBL", "RMBS",
                     "RMCF", "RMCO", "RMD", "RMI", "RMM", "RMMZ", "RMNI", "RMR", "RMSG", "RMT", "RMTI", "RNA", "RNAC",
                     "RNAZ", "RNG", "RNGR", "RNP", "RNR", "RNST", "RNTX", "RNW", "RNXT", "ROAD", "ROCK", "ROG", "ROIV",
                     "ROK", "ROKU", "ROL", "ROLR", "ROMA", "ROOT", "ROP", "ROST", "RPAY", "RPD", "RPID", "RPM", "RPRX",
                     "RPT", "RPTX", "RQI", "RR", "RRBI", "RRC", "RRGB", "RRR", "RRX", "RS", "RSF", "RSG", "RSI", "RSKD",
                     "RSLS", "RSSS", "RSVR", "RSVRW", "RTC", "RTO", "RTX", "RUM", "RUMBW", "RUN", "RUSHA", "RUSHB",
                     "RVLV", "RVMD", "RVMDW", "RVP", "RVPH", "RVPHW", "RVSB", "RVSN", "RVSNW", "RVT", "RVTY", "RVYL",
                     "RWAY", "RWAYL", "RWAYZ", "RWT", "RWTN", "RWTO", "RWTP", "RXO", "RXRX", "RXST", "RXT", "RY",
                     "RYAAY", "RYAM", "RYAN", "RYDE", "RYET", "RYI", "RYN", "RYTM", "RZB", "RZC", "RZLT", "RZLV",
                     "RZLVW", "S", "SA", "SABA", "SABR", "SABS", "SACH", "SAFE", "SAFT", "SAG", "SAGE", "SAGT", "SAH",
                     "SAIA", "SAIC", "SAIH", "SAIHW", "SAIL", "SAJ", "SAM", "SAMG", "SAN", "SANA", "SAND", "SANG",
                     "SANM", "SANW", "SAP", "SAR", "SARO", "SAT", "SATL", "SATS", "SATX", "SAVA", "SAY", "SAZ", "SB",
                     "SBAC", "SBC", "SBCF", "SBCWW", "SBET", "SBEV", "SBFG", "SBFM", "SBFMW", "SBGI", "SBH", "SBI",
                     "SBLK", "SBR", "SBRA", "SBS", "SBSI", "SBSW", "SBUX", "SBXD", "SCCC", "SCCD", "SCCE", "SCCF",
                     "SCCG", "SCCO", "SCD", "SCE", "SCHL", "SCHW", "SCI", "SCKT", "SCL", "SCLX", "SCLXW", "SCM", "SCNI",
                     "SCNX", "SCOR", "SCPH", "SCS", "SCSC", "SCVL", "SCWO", "SCYX", "SD", "SDA", "SDAWW", "SDGR",
                     "SDHC", "SDHIU", "SDHY", "SDOT", "SDRL", "SDST", "SDSTW", "SE", "SEAL", "SEAT", "SEATW", "SEB",
                     "SEDG", "SEE", "SEED", "SEER", "SEG", "SEI", "SEIC", "SELF", "SELX", "SEM", "SEMR", "SENEA",
                     "SENEB", "SENS", "SEPN", "SER", "SERA", "SERV", "SES", "SEVN", "SEZL", "SF", "SFB", "SFBC", "SFBS",
                     "SFD", "SFHG", "SFIX", "SFL", "SFM", "SFNC", "SFST", "SFWL", "SG", "SGA", "SGBX", "SGC", "SGD",
                     "SGHC", "SGHT", "SGI", "SGLY", "SGMA", "SGML", "SGMO", "SGMT", "SGN", "SGRP", "SGRY", "SGU",
                     "SHAK", "SHBI", "SHC", "SHCO", "SHEL", "SHEN", "SHFS", "SHFSW", "SHG", "SHIM", "SHIP", "SHLS",
                     "SHMD", "SHMDW", "SHO", "SHOO", "SHOP", "SHOT", "SHOTW", "SHPH", "SHW", "SHYF", "SIBN", "SID",
                     "SIDU", "SIEB", "SIF", "SIFY", "SIG", "SIGA", "SIGI", "SIGIP", "SII", "SILA", "SILC", "SILO",
                     "SIM", "SIMA", "SIMAU", "SIMAW", "SIMO", "SINT", "SION", "SIRI", "SISI", "SITC", "SITE", "SITM",
                     "SJ", "SJM", "SJT", "SJW", "SKBL", "SKE", "SKIL", "SKIN", "SKK", "SKLZ", "SKM", "SKT", "SKWD",
                     "SKX", "SKY", "SKYE", "SKYH", "SKYQ", "SKYT", "SKYW", "SKYX", "SLAB", "SLB", "SLDB", "SLDP",
                     "SLDPW", "SLE", "SLF", "SLG", "SLGL", "SLGN", "SLI", "SLM", "SLMBP", "SLN", "SLND", "SLNG", "SLNH",
                     "SLNHP", "SLNO", "SLP", "SLQT", "SLRC", "SLRN", "SLRX", "SLS", "SLSN", "SLSR", "SLVM", "SLXN",
                     "SLXNW", "SM", "SMA", "SMBC", "SMBK", "SMC", "SMCI", "SMFG", "SMG", "SMHI", "SMID", "SMLR", "SMMT",
                     "SMP", "SMPL", "SMR", "SMRT", "SMSI", "SMTC", "SMTI", "SMTK", "SMWB", "SMX", "SMXT", "SMXWW", "SN",
                     "SNA", "SNAL", "SNAP", "SNBR", "SNCR", "SNCRL", "SNCY", "SND", "SNDA", "SNDK", "SNDL", "SNDR",
                     "SNDX", "SNES", "SNEX", "SNFCA", "SNGX", "SNN", "SNOA", "SNOW", "SNPS", "SNPX", "SNRE", "SNSE",
                     "SNT", "SNTG", "SNTI", "SNV", "SNWV", "SNX", "SNY", "SNYR", "SO", "SOAR", "SOBO", "SOBR", "SOC",
                     "SOFI", "SOGP", "SOHO", "SOHOB", "SOHOO", "SOHU", "SOJC", "SOJD", "SOJE", "SOJF", "SOL", "SOLV",
                     "SON", "SOND", "SONDW", "SONM", "SONN", "SONO", "SONY", "SOPA", "SOPH", "SOR", "SOS", "SOTK",
                     "SOUN", "SOUNW", "SOWG", "SPAI", "SPB", "SPCB", "SPCE", "SPE", "SPFI", "SPG", "SPGI", "SPH",
                     "SPHL", "SPHR", "SPIR", "SPKL", "SPKLW", "SPLP", "SPMA", "SPMC", "SPNS", "SPNT", "SPOK", "SPOT",
                     "SPPL", "SPR", "SPRB", "SPRC", "SPRO", "SPRU", "SPRY", "SPSC", "SPT", "SPTN", "SPWH", "SPWR",
                     "SPWRW", "SPXC", "SPXX", "SQFT", "SQFTP", "SQFTW", "SQM", "SQNS", "SR", "SRAD", "SRBK", "SRCE",
                     "SRDX", "SRE", "SREA", "SRFM", "SRG", "SRI", "SRL", "SRM", "SRPT", "SRRK", "SRTS", "SRV", "SRZN",
                     "SRZNW", "SSB", "SSBI", "SSBK", "SSD", "SSII", "SSKN", "SSL", "SSNC", "SSP", "SSRM", "SSSS",
                     "SSSSL", "SST", "SSTI", "SSTK", "SSY", "SSYS", "ST", "STAA", "STAG", "STAI", "STAK", "STBA",
                     "STBX", "STC", "STE", "STEC", "STEL", "STEM", "STEP", "STEW", "STFS", "STG", "STGW", "STHO", "STI",
                     "STIM", "STK", "STKH", "STKL", "STKS", "STLA", "STLD", "STM", "STN", "STNE", "STNG", "STOK", "STR",
                     "STRA", "STRF", "STRK", "STRL", "STRM", "STRO", "STRR", "STRRP", "STRS", "STRT", "STRW", "STSS",
                     "STSSW", "STT", "STTK", "STVN", "STWD", "STX", "STXS", "STZ", "SU", "SUGP", "SUI", "SUN", "SUNE",
                     "SUNS", "SUP", "SUPN", "SUPV", "SURG", "SUUN", "SUZ", "SVC", "SVCC", "SVCCU", "SVCCW", "SVCO",
                     "SVII", "SVIIR", "SVIIU", "SVIIW", "SVM", "SVRA", "SVRE", "SVT", "SVV", "SW", "SWAG", "SWAGW",
                     "SWBI", "SWIM", "SWIN", "SWK", "SWKH", "SWKHL", "SWKS", "SWTX", "SWVL", "SWVLW", "SWX", "SWZ",
                     "SXC", "SXI", "SXT", "SXTC", "SXTP", "SXTPW", "SY", "SYBT", "SYBX", "SYF", "SYK", "SYM", "SYNA",
                     "SYNX", "SYPR", "SYRE", "SYT", "SYTA", "SYTAW", "SYY", "SZZLU", "T", "TAC", "TACHU", "TACT",
                     "TAIT", "TAK", "TAL", "TALK", "TALKW", "TALO", "TANH", "TAOP", "TAP", "TARA", "TARS", "TASK",
                     "TATT", "TAVI", "TAVIR", "TAVIU", "TAYD", "TBB", "TBBB", "TBBK", "TBCH", "TBH", "TBI", "TBLA",
                     "TBLD", "TBMCR", "TBN", "TBPH", "TBRG", "TC", "TCBI", "TCBIO", "TCBK", "TCBS", "TCBX", "TCI",
                     "TCMD", "TCOM", "TCPC", "TCRT", "TCRX", "TCX", "TD", "TDAC", "TDACU", "TDC", "TDF", "TDG", "TDOC",
                     "TDS", "TDTH", "TDUP", "TDW", "TDY", "TE", "TEAF", "TEAM", "TECH", "TECK", "TECTP", "TECX", "TEF",
                     "TEI", "TEL", "TELA", "TELO", "TEM", "TEN", "TENB", "TENX", "TEO", "TER", "TERN", "TEVA", "TEX",
                     "TFC", "TFII", "TFIN", "TFINP", "TFPM", "TFSL", "TFX", "TG", "TGB", "TGI", "TGL", "TGLS", "TGNA",
                     "TGS", "TGT", "TGTX", "TH", "THAR", "THC", "THCH", "THFF", "THG", "THM", "THO", "THQ", "THR",
                     "THRD", "THRM", "THRY", "THS", "THTX", "THW", "TIC", "TIGO", "TIGR", "TIL", "TILE", "TIMB", "TIPT",
                     "TIRX", "TISI", "TITN", "TIVC", "TIXT", "TJX", "TK", "TKC", "TKLF", "TKNO", "TKO", "TKR", "TLF",
                     "TLK", "TLN", "TLPH", "TLRY", "TLS", "TLSA", "TLSI", "TLSIW", "TLX", "TLYS", "TM", "TMC", "TMCI",
                     "TMCWW", "TMDE", "TMDX", "TME", "TMHC", "TMO", "TMP", "TMQ", "TMUS", "TNC", "TNDM", "TNET", "TNFA",
                     "TNGX", "TNK", "TNL", "TNMG", "TNON", "TNONW", "TNXP", "TNYA", "TOI", "TOIIW", "TOL", "TOMZ",
                     "TOON", "TOP", "TOPP", "TOPS", "TOPW", "TORO", "TOST", "TOUR", "TOVX", "TOWN", "TOYO", "TPB",
                     "TPC", "TPCS", "TPET", "TPG", "TPGXL", "TPH", "TPIC", "TPL", "TPR", "TPST", "TPTA", "TPVG", "TR",
                     "TRAK", "TRAW", "TRC", "TRDA", "TREE", "TREX", "TRGP", "TRI", "TRIB", "TRIN", "TRINI", "TRINZ",
                     "TRIP", "TRMB", "TRMD", "TRMK", "TRML", "TRN", "TRNO", "TRNR", "TRNS", "TROO", "TROW", "TROX",
                     "TRP", "TRS", "TRSG", "TRST", "TRT", "TRTN", "TRTX", "TRU", "TRUE", "TRUG", "TRUP", "TRV", "TRVG",
                     "TRVI", "TRX", "TS", "TSAT", "TSBK", "TSBX", "TSCO", "TSE", "TSEM", "TSHA", "TSI", "TSLA", "TSLX",
                     "TSM", "TSN", "TSQ", "TSSI", "TSVT", "TT", "TTAM", "TTAN", "TTC", "TTD", "TTE", "TTEC", "TTEK",
                     "TTGT", "TTI", "TTMI", "TTNP", "TTSH", "TTWO", "TU", "TURB", "TURN", "TUSK", "TUYA", "TV", "TVACU",
                     "TVC", "TVE", "TVGN", "TVGNW", "TVRD", "TVTX", "TW", "TWFG", "TWG", "TWI", "TWIN", "TWLO", "TWN",
                     "TWNP", "TWO", "TWST", "TX", "TXG", "TXMD", "TXN", "TXNM", "TXO", "TXRH", "TXT", "TY", "TYG",
                     "TYGO", "TYL", "TYRA", "TZOO", "TZUP", "U", "UA", "UAA", "UAL", "UAMY", "UAN", "UAVS", "UBCP",
                     "UBER", "UBFO", "UBS", "UBSI", "UBX", "UBXG", "UCAR", "UCB", "UCL", "UCTT", "UDMY", "UDR", "UE",
                     "UEC", "UEIC", "UFCS", "UFG", "UFI", "UFPI", "UFPT", "UG", "UGI", "UGP", "UGRO", "UHAL", "UHG",
                     "UHS", "UHT", "UI", "UIS", "UK", "UKOMW", "UL", "ULBI", "ULCC", "ULH", "ULS", "ULTA", "ULY",
                     "UMAC", "UMBF", "UMBFP", "UMC", "UMH", "UNB", "UNCY", "UNF", "UNFI", "UNH", "UNIT", "UNM", "UNMA",
                     "UNP", "UNTY", "UOKA", "UONE", "UONEK", "UP", "UPB", "UPBD", "UPC", "UPLD", "UPS", "UPST", "UPWK",
                     "UPXI", "URBN", "URG", "URGN", "URI", "UROY", "USA", "USAC", "USAR", "USARW", "USAS", "USAU",
                     "USB", "USCB", "USEA", "USEG", "USFD", "USGO", "USGOW", "USIO", "USLM", "USM", "USNA", "USPH",
                     "UTF", "UTG", "UTHR", "UTI", "UTL", "UTMD", "UTSI", "UTZ", "UUU", "UUUU", "UVE", "UVSP", "UVV",
                     "UWMC", "UXIN", "UYSCU", "UZD", "UZE", "UZF", "V", "VABK", "VAC", "VACH", "VACHW", "VAL", "VALE",
                     "VALN", "VALU", "VANI", "VATE", "VBF", "VBNK", "VBTX", "VC", "VCEL", "VCIC", "VCICU", "VCIG",
                     "VCSA", "VCTR", "VCV", "VCYT", "VECO", "VEEA", "VEEE", "VEEV", "VEL", "VENU", "VEON", "VERA",
                     "VERB", "VERI", "VERO", "VERU", "VERV", "VERX", "VET", "VFC", "VFF", "VFL", "VFS", "VFSWW", "VG",
                     "VGAS", "VGASW", "VGI", "VGM", "VGZ", "VHC", "VHI", "VIASP", "VIAV", "VICI", "VICR", "VIGL", "VIK",
                     "VINP", "VIOT", "VIPS", "VIR", "VIRC", "VIRT", "VIST", "VITL", "VIV", "VIVK", "VIVS", "VKI", "VKQ",
                     "VKTX", "VLCN", "VLGEA", "VLN", "VLO", "VLRS", "VLT", "VLTO", "VLY", "VLYPN", "VLYPO", "VLYPP",
                     "VMAR", "VMC", "VMD", "VMEO", "VMI", "VMO", "VNCE", "VNDA", "VNET", "VNO", "VNOM", "VNRX", "VNT",
                     "VOC", "VOD", "VOR", "VOXR", "VOYA", "VPG", "VPV", "VRA", "VRAR", "VRAX", "VRCA", "VRDN", "VRE",
                     "VREX", "VRM", "VRME", "VRMEW", "VRN", "VRNA", "VRNS", "VRNT", "VRRM", "VRSK", "VRSN", "VRT",
                     "VRTS", "VRTX", "VS", "VSA", "VSAT", "VSCO", "VSEC", "VSEE", "VSEEW", "VSH", "VSME", "VST", "VSTA",
                     "VSTE", "VSTEW", "VSTM", "VSTS", "VTAK", "VTEX", "VTGN", "VTLE", "VTMX", "VTN", "VTOL", "VTR",
                     "VTRS", "VTS", "VTSI", "VTVT", "VTYX", "VUZI", "VVOS", "VVPR", "VVR", "VVV", "VVX", "VXRT", "VYGR",
                     "VYNE", "VYX", "VZLA", "W", "WAB", "WABC", "WAFD", "WAFDP", "WAFU", "WAI", "WAL", "WALD", "WASH",
                     "WAT", "WATT", "WAVE", "WAY", "WB", "WBA", "WBD", "WBS", "WBTN", "WBX", "WCC", "WCN", "WCT", "WD",
                     "WDAY", "WDC", "WDFC", "WDH", "WDI", "WDS", "WEA", "WEAV", "WEC", "WELL", "WEN", "WERN", "WES",
                     "WEST", "WETH", "WETO", "WEX", "WEYS", "WF", "WFC", "WFCF", "WFF", "WFG", "WFRD", "WGO", "WGRX",
                     "WGS", "WGSWW", "WH", "WHD", "WHF", "WHFCL", "WHG", "WHLR", "WHLRD", "WHLRL", "WHLRP", "WHR",
                     "WHWK", "WIA", "WILC", "WIMI", "WINA", "WING", "WINT", "WIT", "WIW", "WIX", "WK", "WKC", "WKEY",
                     "WKHS", "WKSP", "WLAC", "WLDN", "WLDS", "WLDSW", "WLFC", "WLGS", "WLK", "WLKP", "WLY", "WLYB",
                     "WM", "WMB", "WMG", "WMK", "WMPN", "WMS", "WMT", "WNC", "WNEB", "WNS", "WNW", "WOK", "WOLF",
                     "WOOF", "WOR", "WORX", "WOW", "WPC", "WPM", "WPP", "WPRT", "WRAP", "WRB", "WRBY", "WRD", "WRLD",
                     "WRN", "WS", "WSBC", "WSBCP", "WSBF", "WSC", "WSFS", "WSM", "WSO", "WSR", "WST", "WT", "WTBA",
                     "WTF", "WTFC", "WTFCM", "WTFCP", "WTI", "WTM", "WTO", "WTRG", "WTS", "WTTR", "WTW", "WU", "WULF",
                     "WVE", "WVVI", "WVVIP", "WW", "WWD", "WWR", "WWW", "WXM", "WY", "WYHG", "WYNN", "WYY", "X", "XAGE",
                     "XAGEW", "XAIR", "XBIO", "XBIT", "XBP", "XCH", "XCUR", "XEL", "XELB", "XENE", "XERS", "XFLT",
                     "XFOR", "XGN", "XHG", "XHLD", "XHR", "XIFR", "XIN", "XLO", "XMTR", "XNCR", "XNET", "XOM", "XOMA",
                     "XOMAO", "XOMAP", "XOS", "XP", "XPEL", "XPER", "XPEV", "XPL", "XPO", "XPOF", "XPON", "XPRO",
                     "XRAY", "XRTX", "XRX", "XTIA", "XTKG", "XTLB", "XTNT", "XWEL", "XXII", "XYF", "XYL", "XYLO", "XYZ",
                     "Y", "YAAS", "YALA", "YCBD", "YELP", "YETI", "YEXT", "YGMZ", "YHC", "YHGJ", "YHNA", "YHNAR", "YI",
                     "YIBO", "YJ", "YMAB", "YMM", "YORW", "YOSH", "YOTA", "YOTAR", "YOU", "YPF", "YQ", "YRD", "YSG",
                     "YSXT", "YTRA", "YUM", "YUMC", "YXT", "YYAI", "YYGH", "Z", "ZAPP", "ZAPPW", "ZBAI", "ZBAO", "ZBH",
                     "ZBIO", "ZBRA", "ZCAR", "ZCARW", "ZCMD", "ZD", "ZDGE", "ZENA", "ZENV", "ZEO", "ZEOWW", "ZEPP",
                     "ZETA", "ZEUS", "ZG", "ZGN", "ZH", "ZI", "ZIM", "ZIMV", "ZION", "ZIONP", "ZIP", "ZJK", "ZJYL",
                     "ZK", "ZKH", "ZKIN", "ZLAB", "ZM", "ZNTL", "ZONE", "ZOOZ", "ZS", "ZSPC", "ZTEK", "ZTO", "ZTR",
                     "ZTS", "ZUMZ", "ZURA", "ZVIA", "ZVRA", "ZVSA", "ZWS", "ZYBT", "ZYME", "ZYXI"]



import datetime
from typing import Any

import polars as pl
import structlog
from exchange_calendars import ExchangeCalendar, get_calendar

from ziplime.assets.services.asset_service import AssetService
from ziplime.data.domain.data_bundle import DataBundle
from ziplime.data.services.data_bundle_source import DataBundleSource
from ziplime.data.services.bundle_registry import BundleRegistry
from ziplime.data.services.bundle_storage import BundleStorage
from ziplime.utils.class_utils import load_class


class BundleService:

    def __init__(self, bundle_registry: BundleRegistry):
        self._bundle_registry = bundle_registry
        self._logger = structlog.get_logger(__name__)

    async def list_bundles(self) -> list[dict[str, Any]]:
        return await self._bundle_registry.list_bundles()

    async def ingest_bundle(self, name: str,
                            bundle_version: str,
                            date_start: datetime.datetime,
                            date_end: datetime.datetime,
                            trading_calendar: ExchangeCalendar,
                            symbols: list[str],
                            data_bundle_source: DataBundleSource,
                            frequency: datetime.timedelta,
                            bundle_storage: BundleStorage,
                            asset_service: AssetService
                            ):

        """Ingest data for a given bundle.        """
        self._logger.info(f"Ingesting bundle: name={name}, date_start={date_start}, date_end={date_end}, "
                          f"symbols={symbols}, frequency={frequency}")
        if date_start < trading_calendar.first_session.replace(tzinfo=trading_calendar.tz):
            raise ValueError(
                f"Date start must be after first session of trading calendar. "
                f"First session is {trading_calendar.first_session.replace(tzinfo=trading_calendar.tz)} "
                f"and date start is {date_start}")

        if date_end > trading_calendar.last_session.replace(tzinfo=trading_calendar.tz):
            raise ValueError(
                f"Date end must be before last session of trading calendar. "
                f"Last session is {trading_calendar.last_session.replace(tzinfo=trading_calendar.tz)} "
                f"and date end is {date_end}")

        data = await data_bundle_source.get_data(
            symbols=symbols,
            frequency=frequency,
            date_from=date_start,
            date_to=date_end
        )
        if data.is_empty():
            self._logger.warning(
                "No data for symbols={symbols}, frequency={frequency}, date_from={date_from}, date_end={date_end} found. Skipping ingestion.")
            return
        data = data.with_columns(
            pl.lit(0).alias("sid"),
            pl.lit(False).alias("backfilled")
        )
        # repair data
        all_bars = [s for s in pl.from_pandas(
            trading_calendar.sessions_minutes(start=date_start.replace(tzinfo=None), end=date_end.replace(tzinfo=None)).tz_convert(trading_calendar.tz)
        ) if s >= date_start and s <= date_end]
        required_sessions = pl.DataFrame({"date": all_bars, "close": 0.00}).group_by_dynamic(
            index_column="date", every=frequency
        ).df

        equities_by_exchange = data.select("symbol", "exchange", "exchange_country").group_by("exchange", "exchange_country").agg(pl.col("symbol").unique())
        for row in equities_by_exchange.iter_rows(named=True):
            exchange_name = row["exchange"]
            exchange_country = row["exchange_country"]
            symbols = row["symbol"]
            equities = await asset_service.get_equities_by_symbols(symbols=symbols, exchange_name=exchange_name)
            symbol_to_sid = {e.get_symbol_by_exchange(exchange_name=exchange_name): e.sid for e in equities}

            for symbol in symbols:
                symbol_data = data.filter(symbol=symbol).with_columns(pl.col("date"))
                missing_sessions = sorted(set(required_sessions["date"]) - set(symbol_data["date"]))
                if len(missing_sessions) > 0:
                    self._logger.warning(
                        f"Data for symbol {symbol} is missing on ticks ({len(missing_sessions)}): {[missing_session.isoformat() for missing_session in missing_sessions]}")
                    # new_rows = {"date": None for col in data.columns}
                    # for missing_session in missing_sessions:
                    #     # Create a dictionary with None for all columns
                    #     new_rows = {col: None for col in data.columns}
                    #     # Set the specified column to the desired value
                    #     new_rows[column_name] = value

                    # Create a DataFrame with just this one row
                    new_rows_df = pl.DataFrame({"date": missing_sessions, "symbol": symbol, "exchange": exchange_name,
                                                "exchange_country": exchange_country},
                                               schema_overrides={"date": data.schema["date"]})

                    # Concatenate with the original DataFrame
                    data = pl.concat([data, new_rows_df], how="diagonal")

            data = data.with_columns(
                pl.col("symbol").replace(symbol_to_sid).cast(pl.Int64).alias("sid")
            ).sort(["sid", "date"])



        # equities = data.group_by("symbol", "exchange").agg(pl.max("date").alias("max_date"),
        #                                                    pl.min("date").alias("min_date"))
        #
        # exchanges = [
        #     ExchangeInfo(exchange=exchange["exchange"],
        #              canonical_name=exchange["exchange"],
        #              country_code="US")  # TODO: make dynamic
        #     for exchange in equities.select(pl.col("exchange")).unique().iter_rows(named=True)
        # ]
        # assets_db = []
        # equity_symbol_mappings = []
        # currency_asset_router = AssetRouter(sid=1, asset_type="currency")
        # asset_routers_db = [currency_asset_router]
        # trading_pairs = []
        #
        # currencies_db = [
        #     CurrencyModel(sid=currency_asset_router.sid,
        #              start_date=datetime.date.today().replace(year=1900),
        #              first_traded=datetime.date.today().replace(year=1900),
        #              end_date=datetime.date.today().replace(year=2099),
        #              asset_name="Currency USD",
        #              symbol="USD",
        #              auto_close_date=datetime.date.today().replace(year=2099), )
        # ]
        # sid_counter = 2
        #
        # for idx, asset in enumerate(equities.iter_rows(named=True)):
        #     # asset_router = AssetRouter(
        #     #     sid=sid_counter,
        #     #     asset_type=AssetType.EQUITY.value
        #     # )
        #     # asset_routers_db.append(asset_router)
        #     asset_db = Equity(
        #         sid=asset_router.sid,
        #         start_date=asset["min_date"].date().replace(year=1900),
        #         first_traded=asset["min_date"].date(),
        #         end_date=asset["max_date"].date().replace(year=2099),
        #         asset_name=asset["symbol"],
        #         auto_close_date=asset["max_date"].date().replace(year=2099),
        #         symbol_mapping={"LIME": asset["symbol"]}
        #     )
        #     assets_db.append(asset_db)
        #     equity_symbol_mapping = EquitySymbolMappingModel(
        #         sid=asset_db.sid,
        #         company_symbol=asset["symbol"],
        #         symbol=asset["symbol"],
        #         share_class_symbol="",
        #         start_date=asset_db.start_date.replace(year=1900),
        #         end_date=asset_db.end_date.replace(year=2099),
        #         exchange=asset["exchange"],
        #     )
        #     equity_symbol_mappings.append(equity_symbol_mapping)
        #
        #     # trading_pair = TradingPairModel(
        #     #     id=uuid.uuid4(),
        #     #     base_asset_sid=asset_router.sid,
        #     #     quote_asset_sid=asset_db.sid,
        #     #     exchange=asset["exchange"],
        #     # )
        #     # trading_pairs.append(trading_pair)
        #
        #     data = data.with_columns(
        #         sid=pl.when(pl.col("symbol") == asset["symbol"]).then(sid_counter).otherwise(data["sid"]))
        #     sid_counter += 1
        #
        # await assets_repository.save_exchanges(exchanges=exchanges)
        # await assets_repository.save_asset_routers(asset_routers=asset_routers_db)
        # await assets_repository.save_currencies(currencies=currencies)
        # await assets_repository.save_equities(equities=assets_db)
        # await assets_repository.save_trading_pairs(trading_pairs=trading_pairs)
        # await assets_repository.save_equity_symbol_mappings(equity_symbol_mappings=equity_symbol_mappings)
        #
        data_bundle = DataBundle(name=name,
                                 start_date=date_start,
                                 end_date=date_end,
                                 trading_calendar=trading_calendar,
                                 frequency=frequency,
                                 data=data,
                                 timestamp=datetime.datetime.now(tz=trading_calendar.tz),
                                 version=bundle_version
                                 )
        await self._bundle_registry.register_bundle(data_bundle=data_bundle, bundle_storage=bundle_storage)
        await bundle_storage.store_bundle(data_bundle=data_bundle)

        self._logger.info(f"Finished ingesting bundle_name={name}, bundle_version={bundle_version}")

    async def load_bundle(self, bundle_name: str, bundle_version: str | None) -> DataBundle:
        bundle_metadata = await self._bundle_registry.load_bundle_metadata(bundle_name=bundle_name,
                                                                           bundle_version=bundle_version)

        bundle_storage_class: BundleStorage = load_class(
            module_name='.'.join(bundle_metadata["bundle_storage_class"].split(".")[:-1]),
            class_name=bundle_metadata["bundle_storage_class"].split(".")[-1])
        # asset_repository_class: AssetRepository = load_class(
        #     module_name='.'.join(bundle_metadata["asset_repository_class"].split(".")[:-1]),
        #     class_name=bundle_metadata["asset_repository_class"].split(".")[-1])
        # adjustment_repository_class: AdjustmentRepository = load_class(
        #     module_name='.'.join(bundle_metadata["adjustment_repository_class"].split(".")[:-1]),
        #     class_name=bundle_metadata["adjustment_repository_class"].split(".")[-1])

        bundle_storage = await bundle_storage_class.from_json(bundle_metadata["bundle_storage_data"])

        # asset_repository = asset_repository_class.from_json(bundle_metadata["asset_repository_data"])
        # adjustment_repository = adjustment_repository_class.from_json(bundle_metadata["adjustment_repository_data"])
        start_date = datetime.datetime.strptime(bundle_metadata["start_date"], "%Y-%m-%dT%H:%M:%SZ")
        trading_calendar = get_calendar(bundle_metadata["trading_calendar_name"],
                                        start=start_date - datetime.timedelta(days=30))
        start_date = start_date.replace(tzinfo=trading_calendar.tz)
        end_date = datetime.datetime.strptime(bundle_metadata["end_date"], "%Y-%m-%dT%H:%M:%SZ").replace(
            tzinfo=trading_calendar.tz)
        frequency = datetime.timedelta(seconds=int(bundle_metadata["frequency_seconds"]))
        timestamp = datetime.datetime.strptime(bundle_metadata["timestamp"], "%Y-%m-%dT%H:%M:%SZ").replace(
            tzinfo=trading_calendar.tz)
        data_bundle = DataBundle(name=bundle_name,
                                 start_date=start_date,
                                 end_date=end_date,
                                 trading_calendar=trading_calendar,
                                 frequency=frequency,
                                 data=None,
                                 timestamp=timestamp,
                                 version=bundle_metadata["version"]
                                 )
        data = await bundle_storage.load_data_bundle(data_bundle=data_bundle)
        data_bundle.data = data
        return data_bundle
        # data_portal = DataPortal(
        #     data_bundle=data_bundle,
        #     historical_data_reader=data_bundle.historical_data_reader,
        #     fundamental_data_reader=data_bundle.fundamental_data_reader,
        #     future_minute_reader=data_bundle.historical_data_reader,
        #     future_daily_reader=data_bundle.historical_data_reader,
        # )
        # return data_portal

    async def clean(self, bundle_name: str, before: datetime.datetime = None, after: datetime.datetime = None,
                    keep_last: bool = None):
        """Clean up data that was created with ``ingest`` or
        ``$ python -m ziplime ingest``

        Parameters
        ----------
        name : str
            The name of the bundle to remove data for.
        before : datetime, optional
            Remove data ingested before this date.
            This argument is mutually exclusive with: keep_last
        after : datetime, optional
            Remove data ingested after this date.
            This argument is mutually exclusive with: keep_last
        keep_last : int, optional
            Remove all but the last ``keep_last`` ingestions.
            This argument is mutually exclusive with:
              before
              after

        Returns
        -------
        cleaned : set[str]
            The names of the runs that were removed.

        Raises
        ------
        BadClean
            Raised when ``before`` and or ``after`` are passed with
            ``keep_last``. This is a subclass of ``ValueError``.
        """

        for bundle in await self._bundle_registry.list_bundles():
            self._delete_bundle(bundle)

        # try:
        #     all_runs = sorted(
        #         filter(
        #             complement(pth.hidden),
        #             os.listdir(pth.data_path([name])),
        #         ),
        #         key=from_bundle_ingest_dirname,
        #     )
        # except OSError as e:
        #     if e.errno != errno.ENOENT:
        #         raise
        #     raise UnknownBundle(name)
        #
        # if before is after is keep_last is None:
        #     raise BadClean(before, after, keep_last)
        # if (before is not None or after is not None) and keep_last is not None:
        #     raise BadClean(before, after, keep_last)
        #
        # if keep_last is None:
        #
        #     def should_clean(name):
        #         dt = from_bundle_ingest_dirname(name)
        #         return (before is not None and dt < before) or (
        #                 after is not None and dt > after
        #         )
        #
        # elif keep_last >= 0:
        #     last_n_dts = set(take(keep_last, reversed(all_runs)))
        #
        #     def should_clean(name):
        #         return name not in last_n_dts
        #
        # else:
        #     raise BadClean(before, after, keep_last)
        #
        # cleaned = set()
        # for run in all_runs:
        #     if should_clean(run):
        #         log.info("Cleaning %s.", run)
        #         path = pth.data_path([name, run])
        #         shutil.rmtree(path)
        #         cleaned.add(path)
        #
        # return cleaned



from dataclasses import dataclass
from decimal import Decimal


@dataclass
class Account:
    """
    The account object tracks information about the trading account. The
    values are updated as the algorithm runs and its keys remain unchanged.
    If connected to a exchange, one can update these values with the trading
    account values as reported by the exchange.
    """

    settled_cash: Decimal
    accrued_interest: Decimal
    buying_power: Decimal
    equity_with_loan: Decimal
    total_positions_value: Decimal
    total_positions_exposure: Decimal
    regt_equity: Decimal
    regt_margin: Decimal
    initial_margin_requirement: Decimal
    maintenance_margin_requirement: Decimal
    available_funds: Decimal
    excess_liquidity: Decimal
    cushion: Decimal
    day_trades_remaining: Decimal
    leverage: Decimal
    net_leverage: Decimal
    net_liquidation: Decimal



import datetime
import os
from pathlib import Path

import uvloop
from exchange_calendars import get_calendar, ExchangeCalendar

from ziplime.assets.domain.ordered_contracts import CHAIN_PREDICATES
from ziplime.assets.entities.currency import Currency
from ziplime.assets.entities.currency_symbol_mapping import CurrencySymbolMapping
from ziplime.assets.entities.equity import Equity
from ziplime.assets.entities.equity_symbol_mapping import EquitySymbolMapping
from ziplime.assets.models.exchange_info import ExchangeInfo
from ziplime.assets.repositories.sqlalchemy_adjustments_repository import SqlAlchemyAdjustmentRepository
from ziplime.assets.repositories.sqlalchemy_asset_repository import SqlAlchemyAssetRepository
from ziplime.assets.services.asset_service import AssetService
from ziplime.constants.stock_symbols import ALL_US_STOCK_SYMBOLS
from ziplime.data.services.bundle_service import BundleService
from ziplime.data.services.file_system_bundle_registry import FileSystemBundleRegistry
from ziplime.data.services.file_system_parquet_bundle_storage import FileSystemParquetBundleStorage
from ziplime.data.services.limex_hub_data_source import LimexHubDataSource


async def add_default_assets(asset_service: AssetService):
    asset_start_date = datetime.datetime(year=1900, month=1, day=1, tzinfo=datetime.timezone.utc)
    asset_end_date = datetime.datetime(year=2099, month=1, day=1, tzinfo=datetime.timezone.utc)

    usd_currency = Currency(
        asset_name="USD",
        symbol_mapping={
            "LIME": CurrencySymbolMapping(
                symbol="USD",
                exchange_name="LIME",
                start_date=asset_start_date,
                end_date=asset_end_date
            )
        },
        sid=None,
        start_date=asset_start_date,
        end_date=asset_end_date,
        auto_close_date=asset_end_date,
        first_traded=asset_start_date
    )

    equities = [
        Equity(
            asset_name=symbol,
            symbol_mapping={
                "LIME": EquitySymbolMapping(
                    symbol=symbol,
                    exchange_name="LIME",
                    start_date=asset_start_date,
                    end_date=asset_end_date,
                    company_symbol="",
                    share_class_symbol=""
                )
            },
            sid=None,
            start_date=asset_start_date,
            end_date=asset_end_date,
            auto_close_date=asset_end_date,
            first_traded=asset_start_date
        ) for symbol in ALL_US_STOCK_SYMBOLS

    ]

    await asset_service.save_exchanges(
        exchanges=[ExchangeInfo(exchange="LIME", canonical_name="LIME", country_code="US")])
    await asset_service.save_currencies([usd_currency])
    await asset_service.save_equities(equities)


async def _ingest_data(
        start_date: datetime.datetime,
        end_date: datetime.datetime,
        trading_calendar: ExchangeCalendar,
        bundle_name: str,
        symbols: list[str],
        frequency: datetime.timedelta,
        bundle_storage_path: str = str(Path(Path.home(), ".ziplime", "data")),

):
    bundle_registry = FileSystemBundleRegistry(base_data_path=bundle_storage_path)
    bundle_service = BundleService(bundle_registry=bundle_registry)
    bundle_storage = FileSystemParquetBundleStorage(base_data_path=bundle_storage_path, compression_level=5)
    data_bundle_source = LimexHubDataSource.from_env()

    bundle_version = str(int(datetime.datetime.now(tz=trading_calendar.tz).timestamp()))

    db_path = str(Path(Path.home(), ".ziplime", "assets.sqlite").absolute())
    if os.path.exists(db_path):
        os.remove(db_path)
    db_url = f"sqlite+aiosqlite:///{db_path}"
    assets_repository = SqlAlchemyAssetRepository(db_url=db_url, future_chain_predicates=CHAIN_PREDICATES)
    adjustments_repository = SqlAlchemyAdjustmentRepository(db_url=db_url)
    asset_service = AssetService(asset_repository=assets_repository, adjustments_repository=adjustments_repository)

    await add_default_assets(asset_service=asset_service)

    await bundle_service.ingest_bundle(
        date_start=start_date.replace(tzinfo=trading_calendar.tz),
        date_end=end_date.replace(tzinfo=trading_calendar.tz),
        bundle_storage=bundle_storage,
        data_bundle_source=data_bundle_source,
        frequency=frequency,
        symbols=symbols,
        name=bundle_name,
        bundle_version=bundle_version,
        trading_calendar=trading_calendar,
        asset_service=asset_service
    )


def ingest_data(start_date: datetime.datetime, end_date: datetime.datetime,
                symbols: list[str],
                trading_calendar: str,
                bundle_name: str,
                data_frequency: datetime.timedelta = datetime.timedelta(minutes=1)):
    calendar = get_calendar(trading_calendar)
    uvloop.run(_ingest_data(start_date=start_date, end_date=end_date, bundle_name=bundle_name, frequency=data_frequency,
                            trading_calendar=calendar, symbols=symbols))
